import os
import warnings
from pathlib import Path
from circleguard import Circleguard
from unittest import TestCase

KEY = os.environ.get("OSU_API_KEY")
if not KEY:
    KEY = input("Enter your api key: ")

RES = Path(__file__).parent / "resources"
# disabled for now
# set_options(loglevel=20)

# what precision we want to guarantee for our tests
DELTA = 0.00001
# osu! only shows ur to two decimals, so we only guarantee precision to there
UR_DELTA = 0.01
# threshold for frametime
FRAMETIME_LIMIT = 13

class CGTestCase(TestCase):
    @classmethod
    def setUpClass(cls, use_cache=True):
        # pass use_cache=False when we need super precise coordinates for tests
        # to work
        cache_path = Path(__file__).parent / "cache.db"
        cache_path = cache_path if use_cache else None
        cls.cg = Circleguard(KEY, db_path=cache_path)

    def setUp(self):
        # prints TestClassName.testMethodName.
        # See https://stackoverflow.com/a/28745033.
        # later edit: disabled this for now, was getting annoying
        # print(self.id())
        # some weird requests warnings about sockets not getting closed;
        # see https://github.com/psf/requests/issues/3912 for more context and
        # https://github.com/biomadeira/pyPDBeREST/commit/71dfe75859a9086
        # b7e415379702cd61ab61fd6e5
        # for implementation
        warnings.filterwarnings(action="ignore",
                message="unclosed",
                category=ResourceWarning)



import cProfile
from pstats import Stats

from circleguard import Circleguard, ReplayPath
from utils import KEY, RES

# TODO not sure how useful this benchmark is tbh, rework or remove

def benchmark_hits_ur(statistic):
    """
    Parameters
    ----------
    statistic: {"hits", "ur", "both"}
        What statistics to calculate
    """
    profiler = cProfile.Profile()


    num_calls = 20
    def benchmark_func():
        cg = Circleguard(KEY)
        replay = ReplayPath(RES / "legit" / "legit-1.osr")
        if statistic in ["hits", "both"]:
            cg.hits(replay)
        if statistic in ["both", "ur"]:
            cg.ur(replay)


    profiler.enable()
    for _ in range(num_calls):
        benchmark_func()

    profiler.disable()

    stats = Stats(profiler)
    print(f"Average time per call with statistic \"{statistic}\": {stats.total_tt / num_calls:3f} ({num_calls} calls)")

if __name__ == "__main__":
    benchmark_hits_ur(statistic="hits")
    benchmark_hits_ur(statistic="ur")
    benchmark_hits_ur(statistic="both")



import abc
import logging
from pathlib import Path
import os
import sqlite3
import random

import osrparse
from osrparse import ReplayEventOsu
import numpy as np
import wtc

from circleguard.mod import Mod
from circleguard.utils import TRACE, KEY_MASK, RatelimitWeight
from circleguard.loader import Loader
from circleguard.span import Span
from circleguard.game_version import GameVersion, NoGameVersion
from circleguard.map_info import MapInfo


class Loadable(abc.ABC):
    """
    Represents one or multiple replays, which have replay data to be loaded
    from some additional source - the osu! api, local cache, or some other
    location.

    Parameters
    ----------
    cache: bool
        Whether to cache the replay data once loaded.
    """
    def __init__(self, cache):
        self.loaded = False
        self.cache = cache

    @abc.abstractmethod
    def load(self, loader, cache):
        """
        Loads the information this loadable needs to become fully loaded.
        Details left to the subclass implementation.

        Parameters
        ----------
        loader: :class:`~circleguard.loader.Loader`
            The loader to load this loadable with. Although subclasses may not
            end up using a :class:`~circleguard.loader.Loader` to load
            themselves (if they don't load anything from the osu api, for
            instance), a loader is still passed regardless.
            Note: ``loader`` may be ``None``. This means that whatever is
            loading the loadable does not have api access and cannot provide a
            loader. If your loadable requires a loader to properly load itself,
            raise an error on a null ``loader``. If your loadable can load
            itself without a ``loader``, proceed as planned and ignore the null
            ``loader``.
        cache: bool
            Whether to cache the replay data once loaded. This argument
            comes from a parent—either a :class:`~.ReplayContainer` or
            :class:`~circleguard.circleguard.Circleguard` itself. Should the
            loadable already have a set ``cache`` value, that should take
            precedence over the option passed in this method, but if the
            loadable has no preference then it should respect the value passed
            here.
        """
        pass

    @abc.abstractmethod
    def __eq__(self, loadable):
        pass


class LoadableContainer(Loadable):
    """
    A Loadable that holds Loadables, which may be ``ReplayContainer``\s or
    ``Replay``\s.

    Parameters
    ----------
    loadables: list[:class:`~.Loadable`]
        The loadables to hold.
    cache: bool
        Whether to cache the loadables once they are loaded. This will be
        overriden by a ``cache`` option set by a :class:`~Loadable` in
        ``loadables``. This only affects child loadables when they do not have
        a ``cache`` option set.

    Notes
    -----
    This class is intended for situations when you have a list of replays and
    replay containers, but no way to separate or distinguish them. If you want
    to get, say, all the replays out of that list (whether they come from
    replay subclasses already in the list, or the replays held by a replay
    container in the list), this loadable container class has the logic to do
    that for you:

    >>> lc = LoadableContainer(mixed_loadable_list)
    >>> replays = lc.all_replays()

    It can also be useful to info load the replay containers in the list,
    without first filtering the list to remove any replay subclasses:

    >>> cg.load_info(lc)
    >>> # all loadable containers in the list are now info loaded
    >>> cg.load(lc)
    >>> # all loadables in the list are now loaded

    You are very unlikely to want to subclass this class. If you want to add a
    new loadable that holds replays, subclass ``ReplayContainer``.
    """

    def __init__(self, loadables, cache=None):
        super().__init__(cache)
        self.loadables = loadables

    def all_replays(self):
        """
        All the :class:`~.Replay`\s in this loadable container.

        Returns
        -------
        list[:class:`~Replay`]
            All the replays in this loadable container.

        Warnings
        --------
        This list may be incomplete if you do not call
        :meth:`~circleguard.circleguard.Circleguard.load_info` on this loadable
        container first, as any replay containers held in this container will
        likely not have references to their replays yet.
        """
        replays = []
        for loadable in self.loadables:
            if isinstance(loadable, ReplayContainer):
                replays += loadable.all_replays()
            else:
                # loadable is a Replay if it's not a ReplayContainer
                replays.append(loadable)
        return replays

    def load(self, loader, cache):
        cascade_cache = cache if self.cache is None else self.cache
        for loadable in self.loadables:
            loadable.load(loader, cascade_cache)

    def load_info(self, loader):
        for loadable in self.loadables:
            if isinstance(loadable, ReplayContainer):
                loadable.load_info(loader)

    def __eq__(self, loadable):
        if not isinstance(loadable, LoadableContainer):
            return False
        return self.all_replays() == loadable.all_replays()

    def __len__(self):
        return len(self.loadables)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.loadables[key.start:key.stop:key.step]
        return self.loadables[key]

    def __iter__(self):
        return iter(self.loadables)


class ReplayContainer(Loadable):
    """
    A Loadable that holds Replay subclasses, and which has an additional state
    between "unloaded" and "loaded" called "info loaded".

    ReplayContainers start unloaded and become info loaded when
    :meth:`~circleguard.circleguard.Circleguard.load_info` is called. They
    become fully loaded when :meth:`~.circleguard.circleguard.Circleguard.load`
    is called (and if this is called when the ReplayContainer is in the unloaded
    state, :meth:`~Loadable.load` will load info first, then load the replays,
    effectively skipping the info loaded state).

    In the unloaded state, the container has no actual Replay objects. It may
    have limited knowledge about their number or type.

    In the info loaded state, the container has references to Replay objects,
    but those Replay objects are unloaded.

    In the loaded state, the Replay objects in the container are loaded.
    """
    def __init__(self, cache):
        super().__init__(cache)
        self.info_loaded = False

    def load(self, loader, cache=None):
        """
        Loads all :class:`~circleguard.loadables.Loadable`\s contained by this
        loadable container.

        Parameters
        ----------
        loader: :class:`~circleguard.loader.Loader`
            The loader to load the :class:`~circleguard.loadables.Loadable`\s
            with.
        """
        if self.loaded:
            return
        cascade_cache = cache if self.cache is None else self.cache
        self.load_info(loader)
        for replay in self.all_replays():
            replay.load(loader, cascade_cache)
        self.loaded = True

    @abc.abstractmethod
    def load_info(self, loader):
        pass

    @abc.abstractmethod
    def all_replays(self):
        """
        Returns all the :class:`~.Replay`\s in this loadable container.

        Warnings
        --------
        If you want an accurate list of :class:`~.Replay`\s in this instance,
        you must call :func:`~circleguard.circleguard.Circleguard.load_info` on
        this instance before
        :func:`~circleguard.loadables.ReplayContainer.all_replays`. Otherwise,
        this instance is not info loaded, and does not have a complete list of
        replays it represents.
        """
        pass

    def __len__(self):
        return len(self.all_replays())

    def __getitem__(self, key):
        replays = self.all_replays()
        if isinstance(key, slice):
            return replays[key.start:key.stop:key.step]
        return replays[key]

    def __iter__(self):
        return iter(self.all_replays())


class Map(ReplayContainer):
    """
    A map's top plays (leaderboard), as seen on the website.

    Parameters
    ----------
    beatmap_id: int
        The map to represent the top plays for.
    span: str or Span
        A comma separated list of ranges of top plays to retrieve.
        ``span="1-3,6,2-4"`` -> replays in the range ``[1,2,3,4,6]``.
    mods: :class:`~circleguard.mod.ModCombination`
        If passed, only represent replays played with this exact mod
        combination. Due to limitations with the api, fuzzy matching is not
        implemented.
        |br|
        This is applied before ``span``. That is, if ``span="1-2"``
        and ``mods=Mod.HD``, the top two ``HD`` plays on the map are
        represented.
    cache: bool
        Whether to cache the replays once they are loaded.
    """
    def __init__(self, beatmap_id, span, mods=None, cache=None):
        super().__init__(cache)
        self.replays = []
        self.beatmap_id = beatmap_id
        self.mods = mods
        self.span = Span(span)

        # TODO remove in core 6.0.0
        self.map_id = beatmap_id

    def load_info(self, loader):
        if self.info_loaded:
            return
        if not loader:
            raise ValueError("A Map cannot be info loaded without api access")
        for score in loader.replay_info(self.beatmap_id, span=self.span,
            mods=self.mods):
            r = ReplayMap(score.beatmap_id, score.user_id, score.mods,
                cache=self.cache, info=score)
            self.replays.append(r)
        self.info_loaded = True

    def all_replays(self):
        return self.replays

    def __eq__(self, loadable):
        if not isinstance(loadable, Map):
            return False
        return (self.beatmap_id == loadable.beatmap_id
                and self.mods == loadable.mods and self.span == loadable.span)

    def __repr__(self):
        return (f"Map(beatmap_id={self.beatmap_id},cache={self.cache},"
            f"mods={self.mods},span={self.span},replays={self.replays},"
            f"loaded={self.loaded})")

    def __str__(self):
        return f"Map {self.beatmap_id}"


class User(ReplayContainer):
    """
    A user's top plays (pp-wise, as seen on the website).

    Parameters
    ----------
    user_id: int
        The user to represent the top plays for.
    span: str or Span
        A comma separated list of ranges of top plays to retrieve.
        ``span="1-3,6,2-4"`` -> replays in the range ``[1,2,3,4,6]``.
    mods: :class:`~circleguard.mod.ModCombination`
        If passed, only represent replays played with this exact mod
        combination. Due to limitations with the api, fuzzy matching is not
        implemented.
        |br|
        This is applied before ``span``. That is, if ``span="1-2"``
        and ``mods=Mod.HD``, the user's top two ``HD`` plays are represented.
    cache: bool
        Whether to cache the replays once they are loaded.
    available_only: bool
        Whether to represent only replays that have replay data available.
        Replays are filtered on this basis after ``mods`` and ``span``
        are applied. True by default.
    """
    def __init__(self, user_id, span, mods=None, cache=None, \
        available_only=True):
        super().__init__(cache)
        self.replays = []
        self.user_id = user_id
        self.span = Span(span)
        self.mods = mods
        self.available_only = available_only

    def load_info(self, loader):
        if self.info_loaded:
            return
        if not loader:
            raise ValueError("A User cannot be info loaded without api access")
        # thanks to api v1 weirdness, depending on the endpoint we use to
        # retrieve the Score model, the username may or may not be present (to
        # be explicit, `get_scores` includes the username, `get_user_best` does
        # not). We guarantee that this attribute is present for `ReplayMap`s
        # but when we pass an override `info` to it here it won't retrieve
        # the username, meaning it gets stuck with a `None` username. To fix
        # this just manually retrieve the username once here and set
        # `info.username` manually.
        # Ideally this attribute would be lazy-loaded in some form so this call
        # isn't hit until required, but doing so would require more complexity
        # than I'm comfortable with for such minor savings (one api call per
        # unique user, since `loader.username` is @lru_cached).
        username = loader.username(self.user_id)
        for info in loader.get_user_best(self.user_id, self.span, self.mods):
            if self.available_only and not info.replay_available:
                continue
            info.username = username
            r = ReplayMap(info.beatmap_id, info.user_id, info.mods, self.cache,
                info=info)
            self.replays.append(r)
        self.info_loaded = True

    def all_replays(self):
        return self.replays

    def __eq__(self, loadable):
        if not isinstance(loadable, User):
            return False
        return (self.user_id == loadable.user_id and self.mods == loadable.mods
                and self.span == loadable.span)

    def __str__(self):
        return f"User {self.user_id}"


class MapUser(ReplayContainer):
    """
    All replays on a map by a user, not just the top replay.

    Parameters
    ----------
    beatmap_id: int
        The beatmap to represent scores by `user_id` on.
    user_id: int
        The user to represent scores on `beatmap_id` for.
    span: str or Span
        A comma separated list of ranges of plays to retrieve.
        ``span="1-3,6,2-4"`` -> replays in the range ``[1,2,3,4,6]``.
    cache: bool
        Whether to cache the replays once they are loaded.
    available_only: bool
        Whether to represent only replays that have replay data available.
        Replays are filtered on this basis after ``span`` is applied.
        True by default.
    """
    def __init__(self, beatmap_id, user_id, span=Loader.MAX_MAP_SPAN, \
        cache=None, available_only=True):
        super().__init__(cache)
        self.replays = []
        self.beatmap_id = beatmap_id
        self.user_id = user_id
        self.span = Span(span)
        self.available_only = available_only

        # TODO remove in core 6.0.0
        self.map_id = beatmap_id

    def load_info(self, loader):
        if self.info_loaded:
            return
        if not loader:
            raise ValueError("A MapUser cannot be info loaded without "
                "api access")
        for info in loader.replay_info(self.beatmap_id, span=self.span,
            user_id=self.user_id, limit=False):
            if self.available_only and not info.replay_available:
                continue
            r = ReplayMap(info.beatmap_id, info.user_id, info.mods, self.cache,
                info=info)
            self.replays.append(r)
        self.info_loaded = True

    def all_replays(self):
        return self.replays

    def __eq__(self, loadable):
        if not isinstance(loadable, MapUser):
            return False
        return (self.beatmap_id == loadable.beatmap_id and
            self.user_id == loadable.user_id and self.span == loadable.span)

    def __str__(self):
        return f"MapUser for {self.user_id} on /b/{self.beatmap_id}"

class ReplayCache(ReplayContainer):
    """
    Contains replays represented by a circlecore database. Primarily useful
    to randomly sample these replays, rather than directly access them.

    Parameters
    ----------
    path: string
        The path to the database to load replays from.
    num_maps: int
        How many (randomly chosen) maps to load replays from.
    limit: int
        How many replays to load for each map.

    Notes
    -----
    :meth:`~circleguard.circleguard.Circleguard.load_info` is an expensive
    operation for large databases created on circlecore version 4.3.5 or
    earlier, as they do not have the necessary indexes.
    |br|
    For databases created in later versions, this is a nonissue and the lookup
    is fast.
    """
    def __init__(self, path, num_maps, num_replays):
        super().__init__(False)
        self.path = path
        self.num_maps = num_maps
        self.limit = num_replays * num_maps
        self.replays = []
        conn = sqlite3.connect(path)
        self.cursor = conn.cursor()

    def load_info(self, loader):
        if self.info_loaded:
            return
        map_ids = self.cursor.execute(
            """
            SELECT DISTINCT map_id
            FROM replays
            """
        ).fetchall()

        # flatten map_ids, because it's actually a list of lists
        map_ids = [item[0] for item in map_ids]
        chosen_maps = random.choices(map_ids, k=self.num_maps)

        subclauses = [f"map_id = {chosen_map}" for chosen_map in chosen_maps]
        where_clause = " OR ".join(subclauses)

        # TODO LIMIT clause isn't quite right here, some maps will have less
        # than ``num_replays`` stored
        infos = self.cursor.execute(
            f"""
            SELECT user_id, map_id, replay_data, replay_id, mods
            FROM replays
            WHERE {where_clause}
            LIMIT {self.limit}
            """
        )

        for info in infos:
            r = CachedReplay(info[0], info[1], info[4], info[2], info[3])
            self.replays.append(r)
        self.info_loaded = True

    def all_replays(self):
        return self.replays

    def __eq__(self, other):
        return self.path == other.path


class ReplayDir(ReplayContainer):
    """
    A folder with replay files inside it.

    Notes
    -----
    Any files not ending in ``.osr`` are ignored.

    Warnings
    --------
    Nested directories are not support (yet). Any folders encountered will be
    ignored.
    """
    def __init__(self, dir_path, cache=None):
        super().__init__(cache)
        self.dir_path = Path(dir_path)
        if not self.dir_path.is_dir():
            raise ValueError(f"Expected path pointing to {self.dir_path} to be"
                " a directory")
        self.replays = []

    def load_info(self, loader):
        if self.info_loaded:
            return
        for path in os.listdir(self.dir_path):
            if not path.endswith(".osr"):
                continue
            replay = ReplayPath(self.dir_path / path)
            self.replays.append(replay)
        self.info_loaded = True

    def all_replays(self):
        return self.replays

    def __eq__(self, other):
        return self.dir_path == other.dir_path


class Replay(Loadable):
    """
    A replay played by a player.

    Parameters
    ----------
    weight: :class:`~circleguard.utils.RatelimitWeight`
        How much it 'costs' to load this replay from the api.
    cache: bool
        Whether to cache this replay once it is loaded.

    Attributes
    ----------
    game_version: :class:`~circleguard.game_version.GameVersion`
        Information about the version of osu! the replay was played on.
    timestamp: :class:`datetime.datetime`
        When the replay was played.
    map_id: int
        The id of the map the replay was played on, or 0 if
        unknown or on an unsubmitted map.
    user_id: int
        The id of the player who played the replay, or 0 if unknown
        (if the player is restricted, for instance). Note that if the
        user id is known, even if the user is restricted, it should still be
        given instead of 0.
    username: str
        The username of the player who played the replay.
    mods: :class:`~circleguard.mod.ModCombination`
        The mods the replay was played with.
    replay_id: int
        The id of the replay, or 0 if the replay is unsubmitted.
    keydowns: ndarray[int]
        The keydowns for each frame of the replay. Keydowns are the keys pressed
        in that frame that were not pressed in the previous frame. See
        :meth:`~circleguard.loadables.Replay.keydowns` for more details.
    t: ndarray[int]
        A 1d array containing the timestamp for each frame.
        |br|
        This is only nonnull after the replay has been loaded.
    xy: ndarray[float]
        A 2d, two column array, containing the ``x`` and ``y`` coordinates of
        each frame in the first and second column respectively.
        |br|
        This is only nonnull after the replay has been loaded.
    k: ndarray[int]
        A 1d array containing the keys pressed for each frame.
        |br|
        This is only nonnull after the replay has been loaded.
    """
    def __init__(self, weight, cache):
        super().__init__(cache)
        self.weight = weight

        # These attributes might or might not be set once the replay loads.
        # Ideally, a replay would provide all of these attributes, but there are
        # some cases when only a subset is available.
        #
        # If only some of these attributes are set after the replay is loaded,
        # some ``Circleguard`` methods may reject this replay, as it does not
        # contain the information necessary to do whatever the method needs to.
        #
        # For instance, if the replay provides ``replay_data`` but not ``mods``,
        # ``Circleguard#similarity`` will reject it, as we will not know whether
        # whether ``Mod.HR`` was enabled on the replay, and thus whether to flip
        # the replay before comparing it to another one.

        # replays have no information about their game version by default.
        # Subclasses might set this if they have more information to provide
        # about their version, whether on instantiation or after being loaded.
        self.game_version = NoGameVersion()
        self.timestamp    = None
        # declared as a property with a getter and setter so we can set
        # map_info's map_id attribute automatically
        self._beatmap_id       = None
        # replays have no information about their map by default.
        # TODO: remove in core 6.0.0, in favor of ``Replay#map_available`` (and
        # possibly other mechanisms).
        self.map_info         = MapInfo()
        self.username         = None
        self.user_id          = None
        self.mods             = None
        self.replay_id        = None
        self.replay_data      = None
        self.replay_hash      = None
        self.count_300        = None
        self.count_100        = None
        self.count_50         = None
        self.count_geki       = None
        self.count_katu       = None
        self.count_miss       = None
        self.score            = None
        self.max_combo        = None
        self.is_perfect_combo = None
        self.life_bar_graph   = None
        self.rng_seed         = None
        self.pp               = None

        # These attributes remain ``None``` when replay is unloaded, or loaded
        # but with no data.
        self.t            = None
        self.xy           = None
        self.k            = None
        self._keydowns    = None

    def beatmap_available(self, _library):
        return bool(self.beatmap_id)

    # TODO remove in core 6.0.0
    map_available = beatmap_available

    def has_data(self):
        """
        Whether this replay has any replay data.

        Returns
        -------
        bool
            Whether this replay has any replay data.

        Notes
        -----
        If this replay is unloaded, it is guaranteed to not have any replay
        data. But if the replay is loaded, it is not guaranteed to have any
        replay data. Some replays do not have any replay data available from
        the api, even after being loaded.
        """
        if not self.loaded:
            return False
        return bool(self.replay_data)

    def beatmap(self, library):
        """
        The beatmap this replay was played on.

        Parameters
        ----------
        library: :class:`slider.library.Library`
            The library used by the calling
            :class:`~circleguard.circleguard.Circleguard` instance. Beatmaps
            which have already been downloaded and are cached in this library
            may be returned here instead of redownloading them.
            |br|
            Beatmaps which we download or create in this method, but were not
            previously stored in the library, may also be stored into the
            library for future use as a result of calling this method.

        Returns
        -------
        :class:`slider.beatmap.Beatmap`
            The beatmap this replay was played on.
        None
            If we do not know what beatmap this replay was played on.
        """
        if not self.beatmap_available(library):
            return None

        return library.lookup_by_id(self.beatmap_id, download=True, save=True)

    def _process_replay_data(self, replay_data):
        """
        Preprocesses the replay data (turns it into numpy arrays) for fast
        manipulation when investigating.

        Paramters
        ---------
        replay_data: list[:class:`~osrparse.Replay.ReplayEvent`]
            A list of :class:`~osrparse.Replay.ReplayEvent` objects,
            representing the actual data of the replay. If the replay could not
            be loaded, this should be ``None``.

        Notes
        -----
        This method must be called before a replay can be considered loaded
        (ie before you set ``loaded`` to ``True``).
        """
        self.replay_data = replay_data
        # replay wasn't available, can't preprocess the data
        if replay_data is None:
            return

        # In rare cases (I'm not quite sure how to reproduce), a replay's replay
        # data can be empty. We check this here to throw a clearer error
        # message than the IndexError we will get shortly after with
        # ``replay_data[0]``.
        #
        # Note that there's an important distinction between ``replay_data``
        # being ``None`` and being the empty list ``[]`` - the former means the
        # api (or osr, or some other source) had no replay data for this replay,
        # and the latter means it *had* replay data, it was just empty replay
        # data.
        #
        # It might be okay to just return as with the ``replay_data is None``
        # case, but I'm erring on the side of caution and throwing for now.
        #
        # See https://github.com/circleguard/circleguard/issues/133 for examples
        # of replays exhibiting this behavior.
        if replay_data == []:
            raise ValueError("This replay's replay data was empty. This should "
                "not happen and is indicative of a misbehaved replay.")

        # TODO we'll want to add proper support for non-std replays at some
        # point, but for now we'll just drop the replay data and early return.
        # This results in identical behavior with previous versions of
        # circlecore, before osrparse supported non-std gamemodes.
        if not isinstance(replay_data[0], ReplayEventOsu):
            self.replay_data = None
            return

        # remove invalid zero time frame at beginning of replay
        # https://github.com/ppy/osu/blob/1587d4b26fbad691242544a62dbf017a78705
        # ae3/osu.Game/Scoring/Legacy/LegacyScoreDecoder.cs#L242-L245
        if replay_data[0].time_delta == 0:
            replay_data = replay_data[1:]

        # t, x, y, k
        data = [[], [], [], []]
        ## TODO try to use a peekable iterator to use an iter for above as well
        # use an iter an an optimization so we don't recreate the list when
        # taking (and removing) the first element
        replay_data = iter(replay_data)
        # The following comments in this method are guesswork, but seems to
        # accurately describe replays. This references the "first" frame
        # assuming that we have already removed the truly first zero time
        # frame, if it is present. So technically the "first" frame below may
        # be the second frame.
        # There are two possibilities for replays:
        # * for replays with a skip in the beginning, the first frame time is
        #   the skip duration. The next frame after that will have a negative
        #   time, to account for the replay data before the skip.
        # * for replays without a skip in the beginning, the first frame time
        #   is -1.
        # Since in the first case the first frame time is a large positive,
        # this would make ``highest_running_t`` large and cause all replay data
        # before the skip to be ignored. To solve this, we initialize
        # ``running_t`` to the first frame's time.
        running_t = next(replay_data).time_delta
        # We consider negative time frames in the middle of replays to be
        # valid, with a caveat. Their negative time is counted toward
        # ``running_t`` (that is, decreases ``running_t``), but any frames
        # after it are ignored, until the total time passed of ignored frames
        # is greater than or equal to the negative frame.
        # There's one more catch - the frame that brings us *out* of this
        # "negative time" section where we're ignoring frames will cause a
        # special frame to be inserted, which has the same time as the frame
        # that brought us *into* the negative time section, and specially
        # calculated x and y positions. Details below.
        # I do not know why stable treats negative time frames in this way.
        # It is not what lazer does, as far as I can tell. But it is the only
        # reasonable explanation for stable behavior. This solution may not,
        # however, be the canonical solution.
        highest_running_t = np.NINF
        # The last positive frame we encountered before entering a negative
        # section.
        last_positive_frame = None
        # the running time when we encountered ``last_positive_frame``. We need
        # to save this as we do not store this information in each individual
        # frame.
        last_positive_frame_cum_time = None
        previous_frame = None
        for e in replay_data:
            # check if we were in a negative section of the play at the
            # previous frame (f0) before applying the current frame (f1), so we
            # can apply special logic if f1 is the frame that gets us out of
            # the negative section.
            was_in_negative_section = running_t < highest_running_t

            e_t = e.time_delta
            running_t += e_t
            highest_running_t = max(highest_running_t, running_t)
            if running_t < highest_running_t:
                # if we weren't in a negative section in f0, f1 is the first
                # frame to bring us into one, so f0 is the last positive frame.
                if not was_in_negative_section:
                    last_positive_frame = previous_frame
                    # we want to set it to the cumulative time before f1
                    # was processed, so subtract out the current e_t
                    last_positive_frame_cum_time = running_t - e_t
                previous_frame = e
                continue

            # if we get here, f1 brought us out of the negative section. In
            # this case, osu! actually inserts a new frame, with:
            # * t = the cumulative time at the last positive frame (yes, this
            #   means there are two frames at the same time in the replay
            #   playback).
            # * x, y = a weighted average between the positions of f0 and f1,
            #   weighted by how close the last positive frame's time is to each
            #   of the two frames' times.
            # * k = the keypresses of the last positive frame.
            if was_in_negative_section:
                data[0].append(last_positive_frame_cum_time)

                # this is [running_t at f0, running_t at f1], to interpolate
                # the last positive frame's time between.
                xp = [running_t - e_t, running_t]

                fp = [previous_frame.x, e.x]
                x = np.interp(last_positive_frame_cum_time, xp, fp)
                data[1].append(x)

                fp = [previous_frame.y, e.y]
                y = np.interp(last_positive_frame_cum_time, xp, fp)
                data[2].append(y)

                data[3].append(last_positive_frame.keys)

            data[0].append(running_t)
            data[1].append(e.x)
            data[2].append(e.y)
            # TODO: are we taking a performance hit here by letting osrparse
            # convert keys to an enum in its replay's init, then converting it
            # back to an int here (since it's faster for us to work with raw
            # ints)?
            # We could add a ``fast_parse`` option to osrparse which doesn't
            # use nice things like enums if this turns out to be a performance
            # issue.
            data[3].append(int(e.keys))
            previous_frame = e

        block = np.array(data)

        t = np.array(block[0], dtype=int)
        xy = np.array([block[1], block[2]], dtype=float).T
        k = np.array(block[3], dtype=int)

        # sort our data by t. Stable so we don't reorder frames with equal
        # times
        t_sort = np.argsort(t, kind="stable")
        t = t[t_sort]
        xy = xy[t_sort]
        k = k[t_sort]

        self.t = t
        self.xy = xy
        self.k = k

    @property
    def beatmap_id(self):
        return self._beatmap_id

    @beatmap_id.setter
    def beatmap_id(self, beatmap_id):
        self._beatmap_id = beatmap_id
        self.map_info.map_id = beatmap_id

    # TODO remove in core 6.0.0
    @property
    def map_id(self):
        return self._beatmap_id

    @map_id.setter
    def map_id(self, beatmap_id):
        self._beatmap_id = beatmap_id
        self.map_info.map_id = beatmap_id

    @property
    def keydowns(self):
        """
        A list of the keys pressed for each frame that were not pressed in the
        previous frame.

        Examples
        --------
        If the first frame (``f1``) has keys ``K1`` and ``f2`` has keys
        ``K1 + K2``, then ``keydowns[1]`` is ``K2``.
        """
        if not self.has_data():
            return None
        # can't do `if not self._keydowns` because the truth value of an
        # ndarray is ambiguous
        if self._keydowns is None:
            keypresses = self.k & KEY_MASK
            self._keydowns = keypresses & ~np.insert(keypresses[:-1], 0, 0)
        return self._keydowns

    def __repr__(self):
        return (f"Replay(timestamp={self.timestamp},"
            f"beatmap_id={self.beatmap_id},user_id={self.user_id},"
            f"mods={self.mods},replay_id={self.replay_id},weight={self.weight},"
            f"loaded={self.loaded},username={self.username})")

    def __str__(self):
        return f"Replay by {self.username} on {self.beatmap_id}"


class ReplayMap(Replay):
    """
    A :class:`~.Replay` that was submitted to online servers.

    Parameters
    ----------
    map_id: int
        The id of the map the replay was played on.
    user_id: int
        The id of the player who played the replay.
    mods: :class:`~circleguard.mod.ModCombination`
        The mods the replay was played with. If ``None``, the
        highest scoring replay of ``user_id`` on ``map_id`` will be loaded,
        regardless of mod combination. Otherwise, the replay with ``mods``
        will be loaded.
    cache: bool
        Whether to cache this replay once it is loaded.

    Notes
    -----
    The following replay-related attributes are available (not ``None``) when
    this replay is unloaded:

    * beatmap_id
    * user_id
    * mods (if passed)

    In addition to the above, the following replay-related attributes are
    available (not ``None``) when this replay is loaded:

    * timestamp
    * username
    * mods
    * replay_id
    * count_300
    * count_100
    * count_50
    * count_geki
    * count_katu
    * count_miss
    * score
    * max_combo
    * is_perfect_combo
    * pp
    * replay_data
    """

    def __init__(self, beatmap_id, user_id, mods=None, cache=None, info=None):
        super().__init__(RatelimitWeight.HEAVY, cache)
        self.log = logging.getLogger(__name__ + ".ReplayMap")
        self.beatmap_id = beatmap_id
        self.user_id = user_id
        self.mods = mods
        self.info = info
        if info:
            self.timestamp = info.date
            self.beatmap_id = info.beatmap_id
            self.user_id = info.user_id
            self.username = info.username
            self.replay_id = info.replay_id
            self.mods = info.mods

        # TODO remove in core 6.0.0
        self.map_id = self.beatmap_id

    def load(self, loader, cache):
        """
        Loads the data for this replay from the api.

        Parameters
        ----------
        loader: :class:`~.loader.Loader`
            The :class:`~.loader.Loader` to load this replay with.
        cache: bool
            Whether to cache this replay after loading it. This only has an
            effect if ``self.cache`` is unset (``None``).
        """
        # only listen to the parent's cache if ours is not set. Lower takes
        # precedence
        cache = cache if self.cache is None else self.cache
        self.log.debug("Loading %r", self)
        if self.loaded:
            self.log.debug("%s already loaded, not loading", self)
            return

        if not loader:
            raise ValueError("A ReplayMap cannot be loaded without api access")

        if self.info:
            info = self.info
        else:
            info = loader.replay_info(self.beatmap_id, user_id=self.user_id,
                mods=self.mods)

        self.timestamp = info.date
        # estimate version with timestamp, this is only accurate if the user
        # keeps their game up to date
        self.game_version = GameVersion.from_datetime(self.timestamp,
            concrete=False)
        self.username = info.username
        self.mods = info.mods
        self.replay_id = info.replay_id
        self.count_100 = info.count_100
        self.count_300 = info.count_300
        self.count_50 = info.count_50
        self.count_geki = info.count_geki
        self.count_katu = info.count_katu
        self.count_miss = info.count_miss
        self.score = info.score
        self.max_combo = info.max_combo
        self.is_perfect_combo = info.perfect
        self.pp = info.pp

        replay_data = loader.replay_data(info, cache=cache)
        self._process_replay_data(replay_data)
        self.loaded = True
        self.log.log(TRACE, "Finished loading %s", self)

    def __eq__(self, loadable):
        """
        Whether the two replay maps are equal.

        Notes
        -----
        This does not take into account the
        ``cache`` attribute, because equality here means "do they represent the
        same replays".
        """
        if not isinstance(loadable, ReplayMap):
            return False
        if self.has_data() and loadable.has_data():
            return self.replay_data == loadable.replay_data
        return (self.beatmap_id == loadable.beatmap_id and
            self.user_id == loadable.user_id and self.mods == loadable.mods)

    def __hash__(self):
        return hash((self.beatmap_id, self.user_id, self.mods))

    def __repr__(self):
        if self.loaded:
            return (f"ReplayMap(timestamp={self.timestamp},"
            f"beatmap_id={self.beatmap_id},user_id={self.user_id},"
            f"mods={self.mods},cache={self.cache},replay_id={self.replay_id},"
            f"loaded={self.loaded},username={self.username})")
        return (f"ReplayMap(beatmap_id={self.beatmap_id},"
            f"user_id={self.user_id},mods={self.mods},cache={self.cache},"
            f"loaded={self.loaded})")

    def __str__(self):
        return (f"{'Loaded' if self.loaded else 'Unloaded'} ReplayMap by "
            f"{self.user_id} on {self.beatmap_id}")


class ReplayDataOSR(Replay):
    """
    A :class:`~.Replay` which has been saved in the osr format.

    Parameters
    ----------
    weight: :class:`~circleguard.utils.RatelimitWeight`
        How much it 'costs' to load this replay from the api.
    cache: bool
        Whether to cache this replay once it is loaded.

    Notes
    -----
    ReplayDataStrings have no replay-related attributes available (not ``None``)
    when they are unloaded.

    The following replay-related attributes are available (not ``None``) when
    this replay is loaded:

    * timestamp
    * beatmap_id
    * username
    * user_id
    * mods
    * replay_id
    * beatmap_hash
    * replay_hash
    * count_300
    * count_100
    * count_50
    * count_geki
    * count_katu
    * count_miss
    * score
    * max_combo
    * is_perfect_combo
    * life_bar_graph (currently unparsed)
    * replay_data
    """
    def __init__(self, ratelimit_weight, cache=None):
        super().__init__(ratelimit_weight, cache)
        self.log = logging.getLogger(__name__ + ".ReplayPath")
        self.beatmap_hash = None

        self._user_id_func = None
        self._user_id = None
        self._beatmap_id_func = None

    def beatmap(self, library):
        if not self.beatmap_available(library):
            return None
        # if we can't load our beatmap_id, fall back to loading from slider.
        if not self.can_load_api_attributes() and self.beatmap_hash:
            return library.lookup_by_md5(self.beatmap_hash)
        return super().beatmap(library)

    def beatmap_available(self, library):
        beatmap_cached = library.beatmap_cached(beatmap_md5=self.beatmap_hash)
        if self.beatmap_hash and beatmap_cached:
            return True
        return super().beatmap_available(library)

    def load_from_osrparse_replay(self, replay, loader, _cache):
        """
        Loads the data for this replay from the already loaded osrparse replay.

        Parameters
        ----------
        loader: :class:`~.loader.Loader`
            The :class:`~.loader.Loader` to load this replay with.
            |br|
            If ``None``, this replay will be unable to retrieve its ``map_id``
            or ``user_id``, but everything else will still be loaded.
        cache: bool
            Whether to cache this replay after loading it. This only has an
            effect if ``self.cache`` is unset (``None``). Note that currently
            we do not cache :class:`~.ReplayPath` regardless of this parameter.
        """
        self.game_version = GameVersion(replay.game_version, concrete=True)
        self.beatmap_hash = replay.beatmap_hash
        self.username = replay.username
        self.replay_hash = replay.replay_hash
        self.count_300 = replay.count_300
        self.count_100 = replay.count_100
        self.count_50 = replay.count_50
        self.count_geki = replay.count_geki
        self.count_katu = replay.count_katu
        self.count_miss = replay.count_miss
        self.score = replay.score
        self.max_combo = replay.max_combo
        self.is_perfect_combo = replay.perfect
        self.mods = Mod(replay.mods.value)
        self.life_bar_graph = replay.life_bar_graph
        self.timestamp = replay.timestamp
        self.replay_id = replay.replay_id
        self.rng_seed = replay.rng_seed

        if loader:
            self._user_id_func = loader.user_id
            self._beatmap_id_func = loader.beatmap_id

        self._process_replay_data(replay.replay_data)
        self.loaded = True
        self.log.log(TRACE, "Finished loading %s", self)

    def load_from_file(self, path, loader, cache):
        replay = osrparse.Replay.from_path(path)
        self.load_from_osrparse_replay(replay, loader, cache)

    def load_from_string(self, replay_data_str, loader, cache):
        replay = osrparse.Replay.from_string(replay_data_str)
        self.load_from_osrparse_replay(replay, loader, cache)


    @property
    def user_id(self):
        if not self.loaded:
            return None
        if not self._user_id_func:
            raise ValueError("The map if of a replay which has been loaded "
                "without a ``Loader`` cannot be retrieved.")
        if not self._user_id:
            self._user_id = self._user_id_func(self.username)
        return self._user_id

    @property
    def beatmap_id(self):
        if not self.loaded:
            return None
        if not self._beatmap_id_func:
            raise ValueError("The map id of a replay which has been loaded "
                "without a ``Loader`` cannot be retrieved. This can happen if "
                "the replay was loaded with a ``KeylessCircleguard``.")
        # property inheritence is a bit nasty. See
        # https://stackoverflow.com/a/37663266 for reference
        if not super().beatmap_id:
            beatmap_id = self._beatmap_id_func(self.beatmap_hash)
            super(ReplayDataOSR, self.__class__).beatmap_id.fset(self,
                beatmap_id)
        return super().beatmap_id

    @beatmap_id.setter
    def beatmap_id(self, beatmap_id):
        super(ReplayDataOSR, self.__class__).beatmap_id.fset(self, beatmap_id)

    # TODO remove in core 6.0.0
    @property
    def map_id(self):
        if not self.loaded:
            return None
        if not self._beatmap_id_func:
            raise ValueError("The map id of a replay which has been loaded "
                "without a ``Loader`` cannot be retrieved. This can happen if "
                "the replay was loaded with a ``KeylessCircleguard``.")
        if not super().beatmap_id:
            beatmap_id = self._beatmap_id_func(self.beatmap_hash)
            super(ReplayDataOSR, self.__class__).beatmap_id.fset(self,
                beatmap_id)
        return super().beatmap_id

    @map_id.setter
    def map_id(self, map_id):
        super(ReplayDataOSR, self.__class__).map_id.fset(self, map_id)

    def can_load_api_attributes(self):
        """
        Whether we can load attributes that are lazy loaded and require api
        calls, such as ``map_id`` or ``user_id``, if requested.
        """
        return bool(self._beatmap_id_func) and bool(self._user_id_func)

    def api_attributes_loaded(self):
        """
        Whether attributes that are lazy loaded and require api calls, such as
        ``map_id`` or ``user_id``, have already been loaded.
        """
        return bool(self._beatmap_id) and bool(self._user_id)


    @user_id.setter
    def user_id(self, user_id):
        self._user_id = user_id


class ReplayPath(ReplayDataOSR):
    """
    A :class:`~.Replay` saved locally in a ``.osr`` file.

    Parameters
    ----------
    path: str or :class:`os.PathLike`
        The path to the replay file.
    cache: bool
        Whether to cache this replay once it is loaded. Note that currently
        we do not cache :class:`~.ReplayPath` regardless of this parameter.

    Notes
    -----
    ReplayPaths have no replay-related attributes available (not ``None``) when
    they are unloaded.

    The following replay-related attributes are available (not ``None``) when
    this replay is loaded:

    * timestamp
    * beatmap_id
    * username
    * user_id
    * mods
    * replay_id
    * beatmap_hash
    * replay_hash
    * count_300
    * count_100
    * count_50
    * count_geki
    * count_katu
    * count_miss
    * score
    * max_combo
    * is_perfect_combo
    * life_bar_graph (currently unparsed)
    * replay_data
    """

    def __init__(self, path, cache=None):
        super().__init__(RatelimitWeight.LIGHT, cache)
        self.log = logging.getLogger(__name__ + ".ReplayPath")
        self.path = Path(path).absolute()
        self.beatmap_hash = None

        self._user_id_func = None
        self._user_id = None
        self._beatmap_id_func = None

    def load(self, loader, cache):
        self.log.debug("Loading ReplayPath %r", self)
        if self.loaded:
            self.log.debug("%s already loaded, not loading", self)
            return

        self.load_from_file(self.path, loader, cache)

    def __eq__(self, loadable):
        """
        Whether these replay paths are equal.

        Notes
        -----
        If one or both replay paths don't have replay data, this checks path
        equality. If both replay paths have replay data, this checks the
        equality of their replay data.
        |br|
        The reason we don't check path after both are loaded is to avoid
        true in situations like this:

        ```
        r1 = ReplayPath("./1.osr")
        cg.load(r1)
        # change the file located at ./1.osr to another osr file
        r2 = ReplayPath("./1.osr")
        cg.load(r2)
        r1 == r2 # should be False, as they have differing replay data
        ```
        """
        if not isinstance(loadable, ReplayPath):
            return False
        if self.has_data() and loadable.has_data():
            return self.replay_data == loadable.replay_data
        return self.path == loadable.path

    def __hash__(self):
        return hash(self.path)

    def __repr__(self):
        if self.loaded:
            api_attrs_string = ","
            # avoid loading these lazy-loaded attributes by accessing them here,
            # unless they're already loaded
            if self.api_attributes_loaded():
                api_attrs_string = (f"beatmap_id={self.beatmap_id},"
                    f"user_id={self.user_id},")
            return (f"ReplayPath(path={self.path},{api_attrs_string}"
                f"mods={self.mods},replay_id={self.replay_id},"
                f"weight={self.weight},loaded={self.loaded},"
                f"username={self.username})")
        return (f"ReplayPath(path={self.path},weight={self.weight},"
                f"loaded={self.loaded})")

    def __str__(self):
        if self.loaded:
            return (f"Loaded ReplayPath by {self.username} on "
                f"{self.beatmap_id} at {self.path}")
        return f"Unloaded ReplayPath at {self.path}"


class ReplayString(ReplayDataOSR):
    """
    A :class:`~.Replay` saved locally in a ``.osr`` file, when the file has
    already been read as a string.

    Parameters
    ----------
    replay_data_str: str
        The contents of the replay file as a string.
    cache: bool
        Whether to cache this replay once it is loaded. Note that currently
        we do not cache :class:`~.ReplayString` regardless of this parameter.

    Notes
    -----
    ReplayPaths have no replay-related attributes available (not ``None``) when
    they are unloaded.

    The following replay-related attributes are available (not ``None``) when
    this replay is loaded:

    * timestamp
    * beatmap_id
    * username
    * user_id
    * mods
    * replay_id
    * beatmap_hash
    * replay_hash
    * count_300
    * count_100
    * count_50
    * count_geki
    * count_katu
    * count_miss
    * score
    * max_combo
    * is_perfect_combo
    * life_bar_graph (currently unparsed)
    * replay_data

    Examples
    --------
    >>> replay_data = open("replay.osr", "rb").read()
    >>> r = ReplayString(replay_data)
    >>> cg.load(r)
    >>> print(cg.ur(r))
    """

    def __init__(self, replay_data_str, cache=None):
        super().__init__(RatelimitWeight.LIGHT, cache)
        self.log = logging.getLogger(__name__ + ".ReplayString")
        self.replay_data_str = replay_data_str

    def load(self, loader, cache):
        if self.loaded:
            return
        self.load_from_string(self.replay_data_str, loader, cache)

    def __eq__(self, loadable):
        if not isinstance(loadable, ReplayString):
            return False
        return self.replay_data_str == loadable.replay_data_str

    def __hash__(self):
        return hash(self.replay_data_str)

    def __repr__(self):
        if self.loaded:
            api_attrs_string = ","
            if self.api_attributes_loaded():
                api_attrs_string = (f"beatmap_id={self.beatmap_id},"
                    f"user_id={self.user_id},")
            return (f"ReplayString(len(replay_data_str)="
                f"{len(self.replay_data_str)},{api_attrs_string}"
                f"mods={self.mods},"
                f"replay_id={self.replay_id},weight={self.weight},"
                f"loaded={self.loaded},username={self.username})")
        return f"ReplayString(len(replay_data_str)={len(self.replay_data_str)})"

    def __str__(self):
        if self.loaded:
            return (f"Loaded ReplayString by {self.username} on "
                f"{self.beatmap_id}")
        return (f"Unloaded ReplayString with {len(self.replay_data_str)} "
            "chars of data")


class ReplayID(Replay):
    """
    A :class:`~.Replay` that was submitted online and is represented by a unique
    replay id.

    Parameters
    ----------
    replay_id: int
        The id of the replay.
    cache: bool
        Whether to cache this replay once it is loaded. Note that we currently
        do not cache ReplayIDs.

    Notes
    -----
    The following replay-related attributes are available (not ``None``) when
    this replay is unloaded:

    * replay_id

    In addition to the above, the following replay-related attributes are
    available (not ``None``) when this replay is loaded:

    * replay_data
    """
    def __init__(self, replay_id, cache=None):
        super().__init__(RatelimitWeight.HEAVY, cache)
        self.replay_id = replay_id

    def load(self, loader, cache):
        if self.loaded:
            return
        if not loader:
            raise ValueError("A ReplayID cannot be loaded without api access")
        # TODO file github issue about loading info from replay id, right now we
        # can literally only load the replay data which isn't that useful
        cache = cache if self.cache is None else self.cache
        replay_data = loader.replay_data_from_id(self.replay_id, cache)
        self._process_replay_data(replay_data)
        self.loaded = True

    def __eq__(self, other):
        return self.replay_id == other.replay_id

    def __hash__(self):
        return hash(self.replay_id)


class CachedReplay(Replay):
    """
    This class is intended to be instantiated from
    :func:`~.ReplayCache.load_info` and should not be instantiated manually.
    """
    def __init__(self, user_id, beatmap_id, mods, replay_data, replay_id):
        super().__init__(RatelimitWeight.NONE, False)
        self.user_id = user_id
        self.beatmap_id = beatmap_id
        self.mods = Mod(mods)
        self.replay_data = replay_data
        self.replay_id = replay_id

        # TODO remove in core 6.0.0
        self.map_id = beatmap_id

    def load(self, loader, cache):
        if self.loaded:
            return
        decompressed = wtc.decompress(self.replay_data)
        replay_data = osrparse.parse_replay_data(decompressed, decoded=True)
        self._process_replay_data(replay_data)
        self.loaded = True

    def __eq__(self, other):
        return self.replay_id == other.replay_id

    def __hash__(self):
        return hash(self.replay_id)

class ReplayOssapi(ReplayDataOSR):
    """
    Converts a :module:`ossapi` replay to a circlecore :class:`~.Replay`.
    Requires ossapi to be installed (you can't get an ossapi replay without
    having ossapi installed anyway).
    """

    def __init__(self, ossapi_replay):
        super().__init__(RatelimitWeight.NONE, False)

        import ossapi
        game_mode_map = {
            ossapi.GameMode.OSU:    osrparse.GameMode.STD,
            ossapi.GameMode.TAIKO:  osrparse.GameMode.TAIKO,
            ossapi.GameMode.CATCH:  osrparse.GameMode.CTB,
            ossapi.GameMode.MANIA:  osrparse.GameMode.MANIA,
        }

        # an ossapi replay is almost identical to an osrparse replay, except
        # it has a different gamemode and mod enum.
        self.osrparse_replay = osrparse.Replay(
            game_mode_map[ossapi_replay.mode],
            ossapi_replay.game_version,
            ossapi_replay.beatmap_hash,
            ossapi_replay.username,
            ossapi_replay.replay_hash,
            ossapi_replay.count_300,
            ossapi_replay.count_100,
            ossapi_replay.count_50,
            ossapi_replay.count_geki,
            ossapi_replay.count_katu,
            ossapi_replay.count_miss,
            ossapi_replay.score,
            ossapi_replay.max_combo,
            ossapi_replay.perfect,
            osrparse.Mod(ossapi_replay.mods.value),
            ossapi_replay.life_bar_graph,
            ossapi_replay.timestamp,
            ossapi_replay.replay_data,
            ossapi_replay.replay_id,
            ossapi_replay.rng_seed,
        )

    def load(self, loader, cache):
        if self.loaded:
            return

        self.load_from_osrparse_replay(self.osrparse_replay, loader, cache)

    def __eq__(self, loadable):
        if not isinstance(loadable, ReplayOssapi):
            return False
        return self.osrparse_replay == loadable.osrparse_replay

    def __hash__(self):
        return hash(self.osrparse_replay)

    def __str__(self):
        if self.loaded:
            return (f"Loaded ReplayOssapi by {self.username} on "
                f"{self.beatmap_id}")
        return (f"Unloaded ReplayOssapi by {len(self.username)} on beatmap "
            f"hash {self.beatmap_hash}")



class Span(set):
    """
    A set of numbers represented by a string, which can include ranges or
    single numbers, separated by a comma.

    Notes
    -----
    Spans can only range from 1 to 100 inclusive.

    Examples
    --------
    >>> Span("1-3,6,2-4")
    {1, 2, 3, 4, 6}
    """

    def __init__(self, data):
        # allow passing as either span or string
        if not isinstance(data, (Span, str)):
            raise ValueError(f"Expected data to be a str or Span, got type "
                f"{type(data)}.")
        if isinstance(data, Span):
            # python allows initializing a set with a set
            super().__init__(data)
        elif isinstance(data, str):
            span_set = self._to_set(data)
            super().__init__(span_set)

        if max(self) > 100:
            raise ValueError("Spans can only range from 1 to 100 inclusive. "
                f"The largest element passed was {max(self)}")

    def _to_set(self, span):
        """
        Converts a span to the set of numbers covered by that span.

        Parameters
        ----------
        span: str
            The span of numbers to convert to a set. A number may occur more
            than once - whether explicitly or in a range - in the span, but
            will only occur once in the returned set.

        Returns
        -------
        set
            The set of numbers described by the ``span``.

        Examples
        --------
        >>> _to_set("1-3,6,2-4")
        {1, 2, 3, 4, 6}
        """
        ret = set()
        for s in span.split(","):
            if "-" in s:
                p = s.split("-")
                l = list(range(int(p[0]), int(p[1]) + 1))
                ret.update(l)
            else:
                ret.add(int(s))
        return ret



from datetime import timedelta

import numpy as np
from scipy import signal
from slider.beatmap import Circle, Slider, Spinner as SliderSpinner

from circleguard.mod import Mod
from circleguard.utils import KEY_MASK
from circleguard import utils
from circleguard.game_version import GameVersion
from circleguard.hitobjects import Hitobject, Spinner
from circleguard.judgment import JudgmentType, Miss, Hit


class Investigations:
    # https://osu.ppy.sh/home/changelog/stable40/20190207.2
    VERSION_SLIDERBUG_FIXED_STABLE = GameVersion(20190207, concrete=True)
    # https://osu.ppy.sh/home/changelog/cuttingedge/20190111
    VERSION_SLIDERBUG_FIXED_CUTTING_EDGE = GameVersion(20190111, concrete=True)

    @staticmethod
    def ur(replay, beatmap, adjusted):
        """
        Calculates the ur of ``replay`` when played against ``beatmap``.

        Parameters
        ----------
        replay: :class:`~.Replay`
            The replay to calculate the ur of.
        beatmap: :class:`slider.beatmap.Beatmap`
            The beatmap to calculate ``replay``'s ur with.
        adjusted: boolean
            Whether to filter outlier hits before calculating ur.
        """
        # TODO cache hits in replay so we don't recalculate hits for both ur
        # and hits / judgments?
        hits = Investigations.hits(replay, beatmap)
        diffs = [hit.error() for hit in hits]
        if adjusted:
            diffs = utils.filter_outliers(diffs)
        return np.std(diffs) * 10

    @staticmethod
    def snaps(replay, max_angle, min_distance, beatmap):
        """
        Calculates the angle between each set of three points (a,b,c) and finds
        points where this angle is extremely acute and neither ``|ab|`` or
        ``|bc|`` are small.

        Parameters
        ----------
        replay: :class:`~.Replay`
            The replay to investigate for aim correction.
        max_angle: float
            Consider only (a,b,c) where ``∠abc < max_angle``
        min_distance: float
            Consider only (a,b,c) where ``|ab| > min_distance`` and
            ``|bc| > min_distance``.
        beatmap: :class:`slider.beatmap.Beatmap`
            If passed, only the snaps that occur on a hitobject in this beatmap
            will be returned.

        Returns
        -------
        list[:class:`~.Snap`]
            Hits where the angle was less than ``max_angle`` and the distance
            was more than ``min_distance``.

        Notes
        -----
        This does not detect correction where multiple datapoints are placed
        at the correction site (which creates a small ``min_distance``).

        Another possible method is to look at the ratio between the angle
        and distance.
        """
        # when we leave mutliple frames with the same time values, they
        # sometimes get detected (falesly) as aim correction.
        # TODO Worth looking into a bit more to see if we can avoid it without
        # removing the frames entirely.
        t, xy = Investigations.remove_duplicate_t(replay.t, replay.xy)
        t = t[1:-1]

        # label three consecutive points (a b c) and the vectors between them
        # (ab, bc, ac)
        a = xy[:-2]
        b = xy[1:-1]
        c = xy[2:]

        ab = b - a
        bc = c - b
        ac = c - a

        # Distance a to b, b to c, and a to c
        AB = np.linalg.norm(ab, axis=1)
        BC = np.linalg.norm(bc, axis=1)
        AC = np.linalg.norm(ac, axis=1)
        # Law of cosines, solve for beta
        # AC^2 = AB^2 + BC^2 - 2 * AB * BC * cos(beta)
        # cos(beta) = -(AC^2 - AB^2 - BC^2) / (2 * AB * BC)
        num = -(AC ** 2 - AB ** 2 - BC ** 2)
        denom = (2 * AB * BC)
        # use true_divide for handling division by zero
        cos_beta = np.true_divide(num, denom, out=np.full_like(num, np.nan),
            where=denom!=0)
        # rounding issues makes cos_beta go out of arccos' domain, so restrict
        # it
        cos_beta = np.clip(cos_beta, -1, 1)

        beta = np.rad2deg(np.arccos(cos_beta))

        min_AB_BC = np.minimum(AB, BC)
        dist_mask = min_AB_BC > min_distance
        # use less to avoid comparing to nan
        angle_mask = np.less(beta, max_angle, where=~np.isnan(beta))
        # datapoints where both distance and angle requirements are met
        mask = dist_mask & angle_mask

        snaps = []
        for (t, xy, b, d) in zip(t[mask], b[mask], beta[mask], min_AB_BC[mask]):
            # can't discard any snaps if we don't know the beatmap, so count all
            # of them
            if not beatmap:
                snaps.append(Snap(t, b, d))
                continue

            hitobj = beatmap.closest_hitobject(timedelta(milliseconds=int(t)))
            hitobj = Hitobject.from_slider_hitobj(hitobj, replay, beatmap)

            # ignore snaps on spinners
            if isinstance(hitobj, Spinner):
                continue

            easy = Mod.EZ in replay.mods
            hard_rock = Mod.HR in replay.mods
            OD = beatmap.od(easy=easy, hard_rock=hard_rock)

            hitwindow = utils.hitwindow(OD)

            # only count snaps that occur inside hitobjects
            inside_hitobj_pos = np.linalg.norm(xy - hitobj.xy) <= hitobj.radius
            inside_hitobj_t = (hitobj.t - hitwindow) < t < (hitobj.t + hitwindow)
            if inside_hitobj_pos and inside_hitobj_t:
                snaps.append(Snap(t, b, d))

        return snaps

    @staticmethod
    def snaps_sam(replay_data, num_jerks, min_jerk):
        """
        Calculates the jerk at each moment in the Replay, counts the number of
        times it exceeds min_jerk and reports a positive if that number is over
        num_jerks. Also reports all suspicious jerks and their timestamps.

        WARNING
        -------
        Unused function. Kept for historical purposes and ease of viewing in
        case we want to switch to this track of aim correction in the future,
        or provide it as an alternative.
        """

        # get all replay data as an array of type [(t, x, y, k)]
        txyk = np.array(replay_data)

        # drop keypresses
        txy = txyk[:, :3]

        # separate time and space
        t = txy[:, 0]
        xy = txy[:, 1:]

        # j_x = (d/dt)^3 x
        # calculated as (d/dT dT/dt)^3 x = (dT/dt)^3 (d/dT)^3 x
        # (d/dT)^3 x = d(d(dx/dT)/dT)/dT
        # (dT/dt)^3 = 1/(dt/dT)^3
        dtdT = np.diff(t)
        d3xy = np.diff(xy, axis=0, n=3)
        # safely calculate the division and replace with zero if the divisor is
        # zero. dtdT is sliced with 2: because differentiating drops one element
        # for each order (slice (n - 1,) to (n - 3,)).
        # d3xy is of shape (n - 3, 2) so dtdT is also reshaped from (n - 3,) to
        # (n - 3, 1) to align the axes.
        jerk = np.divide(d3xy, dtdT[2:, None] ** 3, out=np.zeros_like(d3xy),
            where=dtdT[2:,None]!=0)

        # take the absolute value of the jerk
        jerk = np.linalg.norm(jerk, axis=1)

        # create a mask of where the jerk reaches suspicious values
        anomalous = jerk > min_jerk
        # and retrieve and store the timestamps and the values themself
        timestamps = t[3:][anomalous]
        values = jerk[anomalous]
        # reshape to an array of type [(t, j)]
        jerks = np.vstack((timestamps, values)).T

        # count the anomalies
        ischeat = anomalous.sum() > num_jerks

        return [jerks, ischeat]

    @staticmethod
    def frametime(replay):
        """
        Calculates the median time between the frames of ``replay``.

        Parameters
        ----------
        replay: :class:`~.Replay`
            The replay to calculate the frametime of.

        Notes
        -----
        Median is used instead of mean to lessen the effect of outliers.
        """
        frametimes = Investigations.frametimes(replay)
        return np.median(frametimes)

    @staticmethod
    def frametimes(replay):
        """
        Returns the time between each two consecutive frames in ``replay``.

        Parameters
        ----------
        replay: :class:`~.Replay`
            The replay to get the frametimes of.
        """
        return np.diff(replay.t)

    @staticmethod
    def keydown_frames(replay):
        """
        Get the frames of ``replay`` which had a keydown event, and we should
        consider eligible to hit a hitobject.

        Parameters
        ----------
        replay: :class:`~.Replay`
            The replay to get the keydown frames of.

        Returns
        -------
        ndarray(float, [float, float])
            The keydown frames for the replay. The first float is the time of
            that frame, the second and third floats are the x and y position
            of the cursor at that frame.
        """
        keydown_frames = []

        # the keydowns for each frame. Frames are "keydown" frames if an
        # additional key was pressed from the previous frame. If keys pressed
        # remained the same or decreased (a key previously pressed is no longer
        # pressed) from the previous frame, ``keydowns`` is zero for that frame.
        keydowns = replay.keydowns

        for i, keydown in enumerate(keydowns):
            if keydown != 0:
                keydown_frames.append([replay.t[i], replay.xy[i]])

        # add a duplicate frame when 2 keys are pressed at the same time
        keydowns = keydowns[keydowns != 0]
        i = 0
        for j in np.where(keydowns == KEY_MASK)[0]:
            keydown_frames.insert(j + i + 1, keydown_frames[j + i])
            i += 1

        return keydown_frames

    @staticmethod
    def hits(replay, beatmap):
        judgment = Investigations.judgments(replay, beatmap)
        judgments = [j for j in judgment if isinstance(j, Hit)]
        return judgments

    # TODO add exception for 2b objects (>1 object at the same time) for current
    # version of notelock
    @staticmethod
    def judgments(replay, beatmap):
        """
        Determines the judgments (where hitobjs were hit or missed, and if hit
        then what type of hit - a 300, 100, or 50) of the given ``replay``
        played against the given ``beatmap``.

        Parameters
        ----------
        replay: :class:`~.Replay`
            The replay to determine the judgments of when played against
            ``beatmap``.
        beatmap: :class:`slider.beatmap.Beatmap`
            The beatmap to determine the udgments of when ``replay`` is played
            against it.
        """
        game_version = replay.game_version

        if not game_version.available():
            # if we have no information about the version, assume it was played
            # after sliderbug was fixed.
            sliderbug_fixed = True

        if not game_version.concrete:
            # if we're only estimating the version, assume the replay was played
            # on stable. if we used the cutting edge version instead, we would
            # be incorrectly using logic for sliderbug being fixed for all
            # replays between the cutting edge version and the stable version
            # which were played on stable.
            # This is wrong for cutting edge replays between those two versions
            # which do not have a concrete version, but that's better than being
            # wrong for stable replays between those two versions.
            sliderbug_fixed = game_version >= Investigations.VERSION_SLIDERBUG_FIXED_STABLE
        else:
            sliderbug_fixed = game_version >= Investigations.VERSION_SLIDERBUG_FIXED_CUTTING_EDGE

        easy = Mod.EZ in replay.mods
        hard_rock = Mod.HR in replay.mods
        hitobjs = beatmap.hit_objects(easy=easy, hard_rock=hard_rock)

        OD = beatmap.od(easy=easy, hard_rock=hard_rock)
        CS = beatmap.cs(easy=easy, hard_rock=hard_rock)
        keydowns = Investigations.keydown_frames(replay)

        judgments = []

        (hw_50, hw_100, hw_300) = utils.hitwindows(OD)
        hitradius = utils.hitradius(CS)

        # keep track of the indices of which hitobjs have been hit so far (none
        # have been hit to start). At the end we'll look through this array and
        # if any index is still ``False``, we'll mark that as a miss.
        hitobj_hit = np.zeros(len(hitobjs), dtype=bool)

        hitobj_i = 0
        keydown_i = 0

        while hitobj_i < len(hitobjs) and keydown_i < len(keydowns):
            hitobj = hitobjs[hitobj_i]
            hitobj_t = hitobj.time.total_seconds() * 1000
            hitobj_xy = [hitobj.position.x, hitobj.position.y]

            keydown_t = keydowns[keydown_i][0]
            keydown_xy = keydowns[keydown_i][1]

            if isinstance(hitobj, Circle):
                hitobj_type = 0
                hitobj_end_time = hitobj_t + hw_50
            elif isinstance(hitobj, Slider):
                hitobj_type = 1
                hitobj_end_time = hitobj.end_time.total_seconds() * 1000
            else:
                hitobj_type = 2
                hitobj_end_time = hitobj.end_time.total_seconds() * 1000

            # before sliderbug fix, notelock ended after hitwindow50
            if not sliderbug_fixed:
                notelock_end_time = hitobj_t + hw_50
                # exception for sliders/spinners, where notelock ends after
                # hitobject end time if it's earlier
                if hitobj_type != 0:
                    notelock_end_time = min(notelock_end_time, hitobj_end_time)
            # after sliderbug fix, notelock ends after hitobject end time
            else:
                notelock_end_time = hitobj_end_time
                # apparently notelock was increased by 1ms for circles
                # (from testing)
                if hitobj_type == 0:
                    notelock_end_time += 1


            # can't press on hitobjects before hitwindowmiss
            if keydown_t < hitobj_t - 400:
                keydown_i += 1
                continue

            if keydown_t <= hitobj_t - hw_50:
                # pressing on a circle or slider during hitwindowmiss will cause
                # a miss
                if np.linalg.norm(keydown_xy - hitobj_xy) <= hitradius and hitobj_type != 2:
                    # sliders don't disappear after missing
                    # so we skip to the press_i that is after notelock_end_time
                    if hitobj_type == 1 and sliderbug_fixed:
                        while keydowns[keydown_i][0] < notelock_end_time:
                            keydown_i += 1
                            if keydown_i >= len(keydowns):
                                break
                    else:
                        keydown_i += 1
                    hitobj_i += 1
                # keypress not on object, so we move to the next keypress
                else:
                    keydown_i += 1
            elif keydown_t >= notelock_end_time:
                # can no longer interact with hitobject after notelock_end_time
                # so we move to the next object
                hitobj_i += 1
            else:
                if (keydown_t < hitobj_t + hw_50 and
                    np.linalg.norm(keydown_xy - hitobj_xy) <= hitradius and
                    hitobj_type != 2):

                    # sliderheads are always 300s even if you click early or
                    # late
                    if hitobj_type == 1:
                        hit_type = JudgmentType.Hit300
                    # TODO: should these ranges be inclusive?
                    elif abs(keydown_t - hitobj_t) < hw_300:
                        hit_type = JudgmentType.Hit300
                    elif abs(keydown_t - hitobj_t) < hw_100:
                        hit_type = JudgmentType.Hit100
                    elif abs(keydown_t - hitobj_t) < hw_50:
                        hit_type = JudgmentType.Hit50

                    judgment = Hit(hitobj, keydown_t, keydown_xy,
                        replay, beatmap, hit_type)
                    judgments.append(judgment)
                    hitobj_hit[hitobj_i] = True

                    # sliders don't disappear after clicking
                    # so we skip to the press_i that is after notelock_end_time
                    if hitobj_type == 1 and sliderbug_fixed:
                        while keydowns[keydown_i][0] < notelock_end_time:
                            keydown_i += 1
                            if keydown_i >= len(keydowns):
                                break
                    else:
                        keydown_i += 1
                    hitobj_i += 1
                # keypress not on object, so we move to the next keypress
                else:
                    keydown_i += 1

        # add a Miss for each hitobj that was never hit
        for i, hitobj_hit_ in enumerate(hitobj_hit):
            # ignore if the hitobj is a spinner, we don't calculate judgments
            # for spinners yet
            if not hitobj_hit_ and not isinstance(hitobjs[i], SliderSpinner):
                judgment = Miss(hitobjs[i], replay, beatmap)
                judgments.append(judgment)

        return judgments

    @staticmethod
    def similarity(replay1, replay2, method, num_chunks, mods_unknown):
        """
        Compares two :class:`~.replay.Replay`\s.

        Parameters
        ----------
        replay1: :class:`~.replay.Replay`
            The first replay to compare.
        replay2: :class:`~.replay.Replay`
            The second replay to compare.

        Returns
        -------
        :class:`~.result.ComparisonResult`
            The result of comparing ``replay1`` to ``replay2``.
        """
        # perform preprocessing here as an optimization, so it is not repeated
        # within different comparison algorithms. This will likely need to
        # become more advanced if we add more (and different) algorithms.

        # interpolation breaks when multiple frames have the same time values
        # (which occurs semi frequently in replays). So filter them out
        t1, xy1 = Investigations.remove_duplicate_t(replay1.t, replay1.xy)
        t2, xy2 = Investigations.remove_duplicate_t(replay2.t, replay2.xy)
        xy1, xy2 = Investigations.interpolate(t1, t2, xy1, xy2)
        xy1, xy2 = Investigations.clean(xy1, xy2)

        # kind of a dirty function with all the switching between similarity
        # and correlation, but I'm not sure I can make it any cleaner

        if not replay1.mods or not replay2.mods:
            # first compute with no modifications
            if method == "similarity":
                sim1 = Investigations.compute_similarity(xy1, xy2)
            if method == "correlation":
                sim1 = Investigations.compute_correlation(xy1, xy2, num_chunks)

            # then compute with hr applied to ``replay1``
            xy1[:, 1] = 384 - xy1[:, 1]

            if method == "similarity":
                sim2 = Investigations.compute_similarity(xy1, xy2)
            if method == "correlation":
                sim2 = Investigations.compute_correlation(xy1, xy2, num_chunks)

            if mods_unknown == "best":
                if method == "similarity":
                    return min(sim1, sim2)
                if method == "correlation":
                    return max(sim1, sim2)

            if mods_unknown == "both":
                return (sim1, sim2)

        # flip if one but not both has HR
        if (Mod.HR in replay1.mods) ^ (Mod.HR in replay2.mods):
            xy1[:, 1] = 384 - xy1[:, 1]

        if method == "similarity":
            return Investigations.compute_similarity(xy1, xy2)
        if method == "correlation":
            return Investigations.compute_correlation(xy1, xy2, num_chunks)

    @staticmethod
    def compute_similarity(xy1, xy2):
        """
        Calculates the average distance between two sets of cursor position
        data.

        Parameters
        ----------
        replay1: ndarray
            The first xy data to compare.
        replay2: ndarray
            The second xy data to compare.

        Returns
        -------
        float
            The mean distance between the two datasets.
        """
        # euclidean distance
        distance = xy1 - xy2
        distance = (distance ** 2).sum(axis=1) ** 0.5
        return distance.mean()

    @staticmethod
    def compute_correlation(xy1, xy2, num_chunks):
        xy1 = xy1.T
        xy2 = xy2.T

        # section into chunks, used to reduce the effect of outlier data
        # (eg. cheater inserts replay data during breaks that places them
        # far away from the actual replay)
        horizontal_length = xy1.shape[1] - xy1.shape[1] % num_chunks
        xy1_parts = np.hsplit(xy1[:,:horizontal_length], num_chunks)
        xy2_parts = np.hsplit(xy2[:,:horizontal_length], num_chunks)
        correlations = []
        for (xy1_part, xy2_part) in zip(xy1_parts, xy2_parts):
            xy1_part -= np.mean(xy1_part)
            xy2_part -= np.mean(xy2_part)
            norm = np.std(xy1_part) * np.std(xy2_part) * xy1_part.size
            # matrix of correlations between xy1 and xy2 at different time
            # shifts
            cross_corr_matrix = signal.correlate(xy1_part, xy2_part) / norm

            # pick the maximum correlation, which will probably be at 0
            # time shift, unless the replays have been intentionally shifted in
            # time
            max_corr = np.max(cross_corr_matrix)
            correlations.append(max_corr)
        # take the median of all the chunks to reduce the effect of outliers
        return np.median(correlations)


    @staticmethod
    def remove_duplicate_t(t, data):
        t, t_sort = np.unique(t, return_index=True)
        data = data[t_sort]
        return (t, data)

    @staticmethod
    def interpolate(t1, t2, xy1, xy2):
        """
        Interpolates the xy data of the shorter replay to the longer replay.

        Returns
        -------
        (ndarray, ndarray)
            The interpolated replay data of the first and second replay
            respectively.

        Notes
        -----
        The length of the two returned arrays will be equal. This is a
        (desirous) side effect of interpolating.
        """
        if len(t1) > len(t2):
            xy2x = np.interp(t1, t2, xy2[:, 0])
            xy2y = np.interp(t1, t2, xy2[:, 1])
            xy2 = np.array([xy2x, xy2y]).T
        else:
            xy1x = np.interp(t2, t1, xy1[:, 0])
            xy1y = np.interp(t2, t1, xy1[:, 1])
            xy1 = np.array([xy1x, xy1y]).T

        return (xy1, xy2)

    @staticmethod
    def clean(xy1, xy2):
        """
        Cleans the given xy data to only include indices where both coordinates
        are inside the osu gameplay window (a 512 by 384 osu!pixel window).

        Warnings
        --------
        The length of the two passed arrays must be equal.
        """
        valid = np.all(([0, 0] <= xy1) & (xy1 <= [512, 384]), axis=1) & \
            np.all(([0, 0] <= xy2) & (xy2 <= [512, 384]), axis=1)
        xy1 = xy1[valid]
        xy2 = xy2[valid]
        return (xy1, xy2)


class Snap:
    """
    A suspicious hit in a replay, specifically so because it snaps away from
    the otherwise normal path. Snaps represent the middle frame in a set of
    three replay frames (so for example ``time`` is the time of the middle
    frame).

    Parameters
    ----------
    time: int
        The time value of the middle datapoint, in ms. 0 represents the
        beginning of the replay.
    angle: float
        The angle between the three datapoints.
    distance: float
        ``min(dist_a_b, dist_b_c)`` if ``a``, ``b``, and ``c`` are three
        datapoints with ``b`` being the middle one.
    """
    def __init__(self, time, angle, distance):
        self.time = time
        self.angle = angle
        self.distance = distance

    def __eq__(self, other):
        return (self.time == other.time and self.angle == other.angle
                and self.distance == other.distance)

    def __hash__(self):
        return hash((self.time, self.angle, self.distance))



import base64
import logging
from lzma import LZMAError
from functools import lru_cache
from pathlib import Path
import sqlite3
import wtc

import osrparse
from ossapi import OssapiV1, ReplayUnavailableException

from circleguard.utils import TRACE
from circleguard.span import Span


class NoInfoAvailableException(Exception):
    def __init__(self):
        super().__init__("No info was available from the api for the given "
            "arguments.")

def check_cache(function):
    """
    A decorator that checks if the passed
    :class:`~.ReplayInfo` has its replay cached. If so,
    returns a :class:`~circleguard.loadables.Replay` instance from the cached
    data. Otherwise, calls and returns the `function` as normal.

    Parameters
    ----------
    function: callable
        The function to wrap.

    Notes
    -----
    ``self`` and ``replay_info`` **MUST** be the first and second arguments to
    the function, respectively.

    Returns
    -------
    :class:`~circleguard.loadables.Replay` or Unknown:
        A :class:`~circleguard.loadables.Replay` instance from the cached data
        if it was cached, or the return value of the function if not.
    """

    def wrapper(*args, **kwargs):
        self = args[0]
        replay_info = args[1]

        decompressed_lzma = self._check_cache(replay_info)
        if not decompressed_lzma:
            return function(*args, **kwargs)

        return osrparse.parse_replay_data(decompressed_lzma, decompressed=True)
    return wrapper



class Loader:
    """
    Manages interactions with the osu api, using the :mod:`ossapi` wrapper.

    Parameters
    ----------
    key: str
        A valid api key. Can be created at
        https://osu.ppy.sh/home/account/edit#legacy-api.
    cache_path: str
        The path to the database to use for caching. A new database will be
        created at this location if one doesn't exist already.
        |br|
        If ``None``, no cache will be used or created.

    Notes
    -----
    If the api ratelimits the key, we wait until our ratelimits are refreshed
    and retry the request. Because the api does not provide the time until the
    next refresh (and we do not use exponential backoff or another retry
    strategy), if the key is ratelimited because of an interaction not managed
    by this class, the class may wait more time than necessary for the key to
    refresh.
    """

    # the maximum number of replay info available through the respective api
    # calls. Note that osu! stores at least the top 1000 replays, but does not
    # make these discoverable unless you know the exact user id, map id, and
    # mods of the replay.
    MAX_MAP_SPAN = Span("1-100")
    MAX_USER_SPAN = Span("1-100")

    def __init__(self, key, cache_path=None, write_to_cache=True):
        self.api = OssapiV1(key)
        self.log = logging.getLogger(__name__)

        self._conn = None
        self._cursor = None
        self.write_to_cache = write_to_cache and bool(cache_path)
        self.read_from_cache = bool(cache_path)

        if cache_path:
            cache_path = Path(cache_path)
            if not cache_path.is_file():
                self._create_cache(cache_path)

            self._conn = sqlite3.connect(str(cache_path))
            self._cursor = self._conn.cursor()

    def replay_info(self, beatmap_id, span=None, user_id=None, mods=None, \
        limit=True):
        """
        Retrieves replay infos from a map's leaderboard.

        Parameters
        ----------
        beatmap_id: int
            The map id to retrieve replay info for.
        span: Span
            A comma separated list of ranges of top replays on the map to
            retrieve. ``span="1-3,6,2-4"`` -> replays in the range
            ``[1,2,3,4,6]``.
        user_id: int
            If passed, only retrieve replay info on ``map_id`` for this user.
            Note that this is not necessarily limited to just the user's top
            score on the map. See ``limit``.
        mods: :class:`~.ModCombination`
            If passed, will only retrieve replay infos for scores that were
            played with the given mods.
        limit: bool
            Whether to limit to only one response. Only has an effect if
            ``user_id`` is passed. If ``limit`` is ``True``, will only return
            the top scoring replay info by ``user_id``. If ``False``, will
            return all scores by ``user_id``.

        Returns
        -------
        list[:class:`~.ReplayInfo`]
            The replay infos representing the map's leaderboard.
        :class:`~.ReplayInfo`
            If ``limit`` is ``True`` and ``user_id`` is passed.

        Notes
        -----
        One of ``user_id`` or ``span`` must be passed.

        Raises
        ------
        NoInfoAvailableException
            If there is no info available for the given parameters.
        """

        # we have to define a new variable to hold locals - otherwise when we
        # call it twice inside the dict comprehension, it rebinds to the comp
        # scope and takes on different locals.
        locals_ = locals()
        self.log.log(TRACE, "Loading replay info on map %d with options %s",
            beatmap_id, {k: locals_[k] for k in locals_ if k != 'self'})

        if not (span or user_id):
            raise ValueError("One of user_id or span must be passed, but not "
                "both")
        api_limit = None
        if span:
            api_limit = max(span)
        mods = None if mods is None else mods.value
        scores = self.api.get_scores(beatmap_id, mode=0, limit=api_limit,
            user=user_id, mods=mods)

        if scores == []:
            # The logic below allows us to load eg
            # ``Map(221777, mods=Mod.SO + Mod.PF + Mod.HT)`` or some equally
            # absurd mod combination for which there are no replays, and have
            # that loading not throw ``NoInfoAvailableException``. Instead,
            # the map's replays list will just be empty.
            # However, we only want to apply this if we're loading a map, ie
            # ``span`` has been passed. If ``user_id`` was passed instead, raise
            # the exception as usual.
            if user_id:
                raise NoInfoAvailableException()
            # the osu! api doesn't distinguish between a map not existing, and
            # no scores having been set on that map for a particular mod
            # combination - both are empty responses which will trigger a no
            # info available exception. We need to figure out which case has
            # occurred here to determine if we should raise or not.
            beatmap_response = self.api.get_beatmaps(beatmap_id=beatmap_id)
            # If the beatmap does not exist, this response will be empty.
            if not beatmap_response:
                raise NoInfoAvailableException()
            # else, the empty response is ok.

        if span:
            # important: if we iterated over ``span`` instead, we would change
            # the order of the scores returned, since ``Span`` is an (unordered)
            # set. Iterate over the scores instead, which have a guaranteed
            # order.
            scores = [score for (i, score) in enumerate(scores, 1) if i in span]

        # limit only applies if user_id was set
        return scores[0] if (limit and user_id) else scores


    def get_user_best(self, user_id, span, mods=None):
        """
        Retrieves replay infos from a user's top plays.

        Parameters
        ----------
        user_id: int
            The user id to get best plays of.
        span: Span
            A comma separated list of ranges of top plays to retrieve.
            ``span="1-3,6,2-4"`` -> replays in the range ``[1,2,3,4,6]``.
        mods: :class:`~.ModCombination`
            If passed, will only retrieve replay infos for scores that were
            played with the given mods.

        Returns
        -------
        list[:class:`~.ReplayInfo`]
            The replay infos representing the user's top plays.
        """
        locals_ = locals()
        self.log.log(TRACE, "Loading user best of %s with options %s",
            user_id, {k: locals_[k] for k in locals_ if k != 'self'})

        scores = self.api.get_user_best(user_id, mode=0, limit=max(span))
        if scores == []:
            raise NoInfoAvailableException()
        if mods:
            _scores = []
            for score in scores:
                if score.mods == mods:
                    _scores.append(score)
            scores = _scores

        # remove span indices which would cause an index error because there
        # weren't that many replay infos returned by the api. eg if there
        # were 4 responses, remove any span above 4
        _span = [x for x in span if x <= len(scores)]
        scores = [scores[i-1] for i in _span]
        return scores


    def load_replay_data(self, beatmap_id, user_id, mods=None):
        """
        Retrieves replay data from the api.

        Parameters
        ----------
        beatmap_id: int
            The map the replay was played on.
        user_id: int
            The user that played the replay.
        mods: :class:`~.ModCombination`
            The mods the replay was played with, or ``None`` for the highest
            scoring replay, regardless of mods.

        Returns
        -------
        str
            The lzma-encoded string, decoded from the base 64 api response,
            representing the replay.
        None
            If no replay data was available.

        Notes
        -----
        This is the low level implementation of :func:`~.replay_data`, handling
        the actual api request.
        """

        self.log.log(TRACE, "Requesting replay data by user %d on map %d with "
            "mods %s", user_id, beatmap_id, mods)
        mods = None if mods is None else mods.value
        content = self.api.get_replay(beatmap_id=beatmap_id, user=user_id,
            mods=mods, mode=0)
        return base64.b64decode(content)

    @check_cache
    def replay_data(self, replay_info, cache=None):
        """
        Retrieves replay data from the api, or from the cache if it is already
        cached.

        Parameters
        ----------
        replay_info: :class:`~.ReplayInfo`
            The replay info representing the replay to retrieve.

        Returns
        -------
        list[:class:`osrparse.replay.ReplayEvent`]
            The replay events with attributes ``x``, ``y``,
            ``time_delta``, and ``keys``.
        None
            If no replay data was available.

        Raises
        ------
        ReplayUnavailableException
            If ``user_info.replay_available` was 1, but we did not receive
            replay data from the api.
        """

        user_id = replay_info.user_id
        beatmap_id = replay_info.beatmap_id
        mods = replay_info.mods
        if not replay_info.replay_available:
            self.log.debug("Replay data by user %d on map %d with mods %s not "
                "available", user_id, beatmap_id, mods)
            return None

        lzma_bytes = self.load_replay_data(beatmap_id, user_id, mods)
        # TODO can this ever be `None`? shouldn't the `base64.b64decode` call in
        # `self.load_replay_data` error on a `None` value? in other words, I
        # don't see how the decode function could ever return `None`.
        if lzma_bytes is None:
            raise ReplayUnavailableException("The api guaranteed there "
                "would be a replay available, but we did not receive any data.")
        try:
            replay_data = osrparse.parse_replay_data(lzma_bytes, decoded=True)
        # see https://github.com/circleguard/circlecore/issues/61
        # api sometimes returns corrupt replays
        except LZMAError:
            self.log.warning("lzma from %r could not be decompressed, api "
                "returned corrupt replay", replay_info)
            return None
        if cache:
            self._cache(lzma_bytes, replay_info)
        return replay_data

    # TODO make this check cache for the replay
    def replay_data_from_id(self, replay_id, _cache):
        """
        Retrieves replay data from the api, given a replay id.

        Parameters
        ----------
        replay_id: int
            The id of the replay to retrieve data for.
        """
        content = self.api.get_replay(score_id=replay_id)
        replay_data = osrparse.parse_replay_data(content)
        # TODO cache the replay here, might require some restructuring/double
        # checking everything will work because we only have its id, not map
        # or user id. In fact I think our db asserts map and user id are nonull
        # so insertion into old dbs probably won't work (and we'd have to change
        # the schema).
        # TODO include a db version in the db for future scenarios like this?
        # look into how that's typically done, maybe just a `VERSION` table with
        # a single row
        return replay_data

    @lru_cache()
    def beatmap_id(self, beatmap_hash):
        """
        Retrieves a beatmap id from a corresponding beatmap hash through the
        api.

        Parameters
        ----------
        map_hash: str
            The md5 hash of the map to get the id of.

        Returns
        -------
        int
            The map id that corresponds to ``map_hash``, or ``0`` if
            ``map_hash`` doesn't mach any map.

        Notes
        -----
        This function is wrapped in a :func:`functools.lru_cache` to prevent
        duplicate api calls.
        """

        beatmaps = self.api.get_beatmaps(beatmap_hash=beatmap_hash)
        if beatmaps == []:
            return 0
        return beatmaps[0].beatmap_id

    # TODO remove in core 6.0.0
    map_id = beatmap_id

    @lru_cache()
    def user_id(self, username):
        """
        Retrieves a user id from a corresponding username through the api.

        Parameters
        ----------
        username: str
            The username of the user to get the user id of.

        Returns
        -------
        int
            The user id that corresponds to ``username``, or ``0`` if
            ``username`` doesn't match any user.

        Notes
        -----
        The api redirects name changes to the current username. For instance,
        ``user_id("cookiezi")`` will return ``124493``, despite shige's current
        osu! username being ``chocomint``. However, I am not sure if this
        behavior is as well defined when someone else takes the previous name
        of a user.

        This function is case insensitive.

        This function is wrapped in a :func:`functools.lru_cache` to prevent
        duplicate api calls.
        """

        user = self.api.get_user(username, user_type="string")
        if user == []:
            return 0
        return user.user_id

    @lru_cache()
    def username(self, user_id):
        """
        Retrieves the username from a corresponding user id through the api.

        Parameters
        ----------
        user_id: int
            The user id of the user to get the username of.

        Returns
        -------
        str
            The username that corresponds to ``user_id``, or an empty string
            if ``user_id`` doesn't match any user.

        Notes
        -----
        This function is the inverse of
        :meth:`~circleguard.loader.Loader.user_id`.

        This function is wrapped in a :func:`functools.lru_cache` to prevent
        duplicate api calls.
        """
        user = self.api.get_user(user_id, user_type="id")
        if user == []:
            return ""
        return user.username

    def _create_cache(self, path):
        """
        Creates a database with the necessary tables at the given path.

        Parameters
        ----------
        path: str
            The absolute path to where the database should be created.

        Notes
        -----
        This function will create directories specified in the path if they
        don't already exist.
        """
        self.log.info("Cache not found at path %s, creating cache", path)
        # create dir if nonexistent
        import os
        if not os.path.exists(path.parent):
            os.makedirs(path.parent)
        conn = sqlite3.connect(str(path))
        c = conn.cursor()
        c.execute(
            """
            CREATE TABLE "REPLAYS" (
                `MAP_ID` INTEGER NOT NULL,
                `USER_ID` INTEGER NOT NULL,
                `REPLAY_DATA` MEDIUMTEXT NOT NULL,
                `REPLAY_ID` INTEGER NOT NULL,
                `MODS` INTEGER NOT NULL,
                PRIMARY KEY(`REPLAY_ID`)
            )""")
        # create our index - this does unfortunately add some size (and
        # insertion time) to the db, but it's worth it to get fast lookups on
        # a map, user, or mods, which are all common operations.
        c.execute(
            """
            CREATE INDEX `lookup_index` ON `REPLAYS` (
                `MAP_ID`, `USER_ID`, `MODS`
            )
            """)
        conn.close()

    def _cache(self, lzma_bytes, replay_info):
        """
        Compresses and caches the given lzma_bytes to the database, linking it
        to the given replay_info. If an entry with the given replay info already
        exists, it is overwritten.

        Parameters
        ----------
        lzma_bytes: str
            The lzma stream to compress and insert into the db.
        replay_info: :class:`~circleguard.loader.ReplayInfo`
            The ReplayInfo object representing this replay.
        """
        if not self.write_to_cache:
            return

        compressed_bytes = wtc.compress(lzma_bytes)
        beatmap_id = replay_info.beatmap_id
        user_id = replay_info.user_id
        mods = replay_info.mods.value
        replay_id = replay_info.replay_id

        self.log.log(TRACE, "Writing compressed lzma to db")
        self._cursor.execute("INSERT INTO replays VALUES(?, ?, ?, ?, ?)",
            [beatmap_id, user_id, compressed_bytes, replay_id, mods])
        self._conn.commit()

    def _check_cache(self, replay_info):
        """
        Checks the cache for a replay matching ``replay_info``.

        Parameters
        ----------
        replay_info: :class:`~circleguard.loader.ReplayInfo`
            The replay info to search for a matching replay with.

        Returns
        -------
        str or None
            The replay data in decompressed lzma form if the cache contains the
            replay, or None if not.
        """
        if not self.read_from_cache:
            return None

        replay_id = replay_info.replay_id

        self.log.log(TRACE, "Checking cache for replay info %s", replay_info)
        result = self._cursor.execute("SELECT replay_data FROM replays WHERE "
            "replay_id=?", [replay_id]).fetchone()
        if result:
            self.log.debug("Loading replay for replay info %s from cache",
                replay_info)
            return wtc.decompress(result[0], decompressed_lzma=True)
        self.log.log(TRACE, "No replay found in cache")



import logging

from circleguard.circleguard import Circleguard, KeylessCircleguard, set_options
from circleguard.loadables import (Replay, ReplayMap, ReplayPath, Map, User,
        MapUser, ReplayDir, ReplayContainer, Loadable, ReplayCache,
        CachedReplay, ReplayID, ReplayString, LoadableContainer, ReplayOssapi)
from circleguard.mod import Mod
from circleguard.loader import Loader, NoInfoAvailableException
from circleguard.version import __version__
from circleguard.investigations import Snap
from circleguard.judgment import Judgment, Hit, Miss, JudgmentType
from circleguard.span import Span
from circleguard.utils import (convert_statistic, order, Key,
        RatelimitWeight, TRACE, ColoredFormatter, replay_pairs, fuzzy_mods,
        hitwindow, hitwindows, hitradius)
from circleguard.game_version import GameVersion, NoGameVersion
from circleguard.hitobjects import Hitobject, Circle, Slider, Spinner

from ossapi.ossapi import (APIException, InvalidKeyException,
        ReplayUnavailableException)

logging.addLevelName(TRACE, "TRACE")
formatter = ColoredFormatter("[%(threadName)s][%(name)s][%(levelname)s]  %(message)s  (%(filename)s:%(lineno)s)")
handler_stream = logging.StreamHandler()
handler_stream.setFormatter(formatter)
logging.getLogger("circleguard").addHandler(handler_stream)

# don't expose ColoredFormatter to consumers
del ColoredFormatter

__all__ = [
# core
"Circleguard", "KeylessCircleguard", "set_options",
# loadables
"ReplayContainer", "Map", "User", "MapUser",
"ReplayCache", "Replay", "ReplayMap", "ReplayPath", "CachedReplay", "Loadable",
"ReplayID", "ReplayDir", "ReplayString", "LoadableContainer", "ReplayOssapi",
# enums
"Key", "RatelimitWeight",
# mod
"Mod",
# utils
"convert_statistic", "order", "Key", "RatelimitWeight", "TRACE", "replay_pairs",
"fuzzy_mods", "hitwindow", "hitwindows", "hitradius",
# loader
"Loader",
# exceptions
"APIException", "NoInfoAvailableException", "InvalidKeyException",
"ReplayUnavailableException",
# version
"__version__",
# investigation-related classes
"Snap", "Judgment", "Hit", "Miss", "JudgmentType",
# span
"Span",
# GameVersion
"GameVersion", "NoGameVersion",
# hitobjects
"Hitobject", "Circle", "Slider", "Spinner"
]



import logging

import wtc
from ossapi import OssapiV1
try:
    import psycopg2
except ImportError:
    raise ImportError("psycopg2 must be installed to use PostgresLoader")

from circleguard.loader import Loader
from circleguard.utils import TRACE
from circleguard.circleguard import Circleguard


class PostgresCircleguard(Circleguard):
    """
    A postgres variant of the default sqlite-backed circleguard.
    """
    def __init__(self, key, db_username, db_password, db_host, db_port, db_name,
        write_to_cache=True, slider_dir=None
    ):
        loader = PostgresLoader(key, db_username, db_password, db_host, db_port,
            db_name, write_to_cache)
        super().__init__(key, loader=loader, slider_dir=slider_dir)

class PostgresLoader(Loader):
    """
    A postgres variant of the default sqlite-backed loader.
    """
    def __init__(
        self, key, db_username, db_password, db_host, db_port, db_name,
        write_to_cache=True
    ):
        self.api = OssapiV1(key)
        self.log = logging.getLogger(__name__)

        self._conn = None
        self._cursor = None
        self.write_to_cache = write_to_cache
        self.read_from_cache = True

        self._conn = psycopg2.connect(
            user=db_username,
            password=db_password,
            host=db_host,
            port=db_port,
            database=db_name
        )
        self._cursor = self._conn.cursor()

    def _check_cache(self, replay_info):
        """
        Checks the cache for a replay matching ``replay_info``.

        Parameters
        ----------
        replay_info: :class:`~circleguard.loader.ReplayInfo`
            The replay info to search for a matching replay with.

        Returns
        -------
        str or None
            The replay data in decompressed lzma form if the cache contains the
            replay, or None if not.
        """
        if not self.read_from_cache:
            return None

        replay_id = replay_info.replay_id

        self.log.log(TRACE, "Checking cache for replay info %s", replay_info)
        self._cursor.execute("SELECT replay_data FROM replays WHERE "
            "replay_id= %s", [replay_id])
        result = self._cursor.fetchone()
        if result:
            self.log.debug("Loading replay for replay info %s from cache",
                replay_info)
            return wtc.decompress(result[0], decompressed_lzma=True)
        self.log.log(TRACE, "No replay found in cache")

    def _cache(self, lzma_bytes, replay_info):
        """
        Compresses and caches the given lzma_bytes to the database, linking it
        to the given replay_info. If an entry with the given replay info already
        exists, it is overwritten.

        Parameters
        ----------
        lzma_bytes: str
            The lzma stream to compress and insert into the db.
        replay_info: :class:`~circleguard.loader.ReplayInfo`
            The ReplayInfo object representing this replay.
        """
        if not self.write_to_cache:
            return

        compressed_bytes = wtc.compress(lzma_bytes)
        beatmap_id = replay_info.beatmap_id
        user_id = replay_info.user_id
        mods = replay_info.mods.value
        replay_id = replay_info.replay_id

        self.log.log(TRACE, "Writing compressed lzma to db")
        self._cursor.execute("INSERT INTO replays VALUES(%s, %s, %s, %s, %s)",
            [replay_id, beatmap_id, user_id, compressed_bytes, mods])
        self._conn.commit()



from datetime import datetime, timezone

from circleguard import (ReplayMap, ReplayPath, RatelimitWeight, Map, User,
    MapUser, Mod, NoInfoAvailableException, ReplayString)

from tests.utils import CGTestCase, RES


class TestReplays(CGTestCase):

    def test_loading_replaypath(self):
        r = ReplayPath(RES / "example_replay.osr")
        self.assertFalse(r.loaded)
        self.cg.load(r)
        self.assertEqual(r.mods, Mod.HD + Mod.DT)
        self.assertEqual(r.replay_id, 2029801532)
        self.assertEqual(r.username, "MarthXT")
        self.assertEqual(r.user_id, 2909663)
        self.assertEqual(r.weight, RatelimitWeight.LIGHT)
        self.assertTrue(r.loaded)
        self.assertEqual(r.beatmap_hash, "c7f9bc1fea826c0f371db08bc5ebc1cc")
        self.assertEqual(r.replay_hash, "266bc8a5f6e9ac0557862da6760388ef")
        self.assertEqual(r.count_300, 154)
        self.assertEqual(r.count_100, 0)
        self.assertEqual(r.count_50, 0)
        self.assertEqual(r.count_geki, 23)
        self.assertEqual(r.count_katu, 0)
        self.assertEqual(r.count_miss, 0)
        self.assertEqual(r.score, 1083482)
        self.assertEqual(r.max_combo, 186)
        self.assertTrue(r.is_perfect_combo)
        self.assertEqual(r.life_bar_graph, None)
        self.assertEqual(r.timestamp, datetime(2015, 12, 16, 19, 40, 39,
            tzinfo=timezone.utc))

    def test_loading_replaymap(self):
        # Toy HDHR score on Pretender
        r = ReplayMap(221777, 2757689)
        self.assertFalse(r.loaded)
        self.cg.load(r)
        self.assertEqual(r.map_id, 221777)
        self.assertEqual(r.user_id, 2757689)
        self.assertEqual(r.mods, Mod.HD + Mod.HR)
        self.assertEqual(r.replay_id, 2832574010)
        self.assertEqual(r.weight, RatelimitWeight.HEAVY)
        self.assertEqual(r.username, "Toy")
        self.assertTrue(r.loaded)
        self.assertEqual(r.count_300, 1449)
        self.assertEqual(r.count_100, 1)
        self.assertEqual(r.count_50, 0)
        self.assertEqual(r.count_geki, 339)
        self.assertEqual(r.count_katu, 1)
        self.assertEqual(r.count_miss, 0)
        self.assertEqual(r.score, 89927731)
        self.assertEqual(r.max_combo, 2388)
        self.assertTrue(r.is_perfect_combo)
        self.assertEqual(r.timestamp, datetime(2019, 6, 19, 3, 22, 44,
            tzinfo=timezone.utc))

    def test_no_replay_raises(self):
        # contrary to loading a Map, where we don't want to raise if the map
        # exists but no scores with the given mod combo exists, we do want to
        # raise if a replay is not available.
        r = ReplayMap(234378, 13947937)
        self.assertRaises(NoInfoAvailableException, lambda: self.cg.load(r))

    def test_no_replay_data_raises(self):
        r = ReplayPath(RES / "other" / "empty_replay_data.osr")
        self.assertRaises(ValueError, lambda: self.cg.load(r))


class TestMap(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.map = Map(221777, "3-5")

    def test_map_load(self):
        self.assertEqual(len(self.map.all_replays()), 0)
        self.assertEqual(len(self.map[:]), 0)
        self.assertFalse(self.map.loaded)
        self.assertFalse(self.map.info_loaded)

        self.cg.load_info(self.map)
        self.assertFalse(self.map.loaded)
        self.assertTrue(self.map.info_loaded)
        self.assertEqual(len(self.map.all_replays()), 3)
        self.assertEqual(len(self.map[:]), 3)

        self.cg.load(self.map)
        self.assertTrue(self.map.loaded)
        self.assertTrue(self.map.info_loaded)

    def test_map_slice(self):
        # sanity check (map id better be what we put in)
        self.assertEqual(self.map[0].map_id, 221777)
        # 4rd (kirby mix)
        self.assertEqual(self.map[1].user_id, 9665206)
        # 3rd, 4th, and 5th (toy, kirby mix, rohulk)
        self.assertListEqual([r.user_id for r in self.map[0:3]], [2757689, 9665206, 3219026])
        # 3rd and 5th (toy and rohulk)
        self.assertListEqual([r.user_id for r in self.map[0:3:2]], [2757689, 3219026])

    def test_no_replays_does_not_raise(self):
        # previously, loading the info of a map or user with no scores on the
        # specified mod combination would result in a NoInfoAvailableException
        # being thrown. We want to make sure this doesn't happen.
        m = Map(2245774, "1-2", Mod.NC + Mod.HR + Mod.SO)
        self.cg.load_info(m)
        self.assertEqual(len(m), 0)


class TestUser(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User(124493, span="1-3")
        cls.tybug = cls.cg.User(12092800, "1-10")

    def test_user_load(self):
        self.assertEqual(len(self.user.all_replays()), 0)
        self.assertEqual(len(self.user[:]), 0)
        self.assertFalse(self.user.loaded)
        self.assertFalse(self.user.info_loaded)

        self.cg.load_info(self.user)
        self.assertFalse(self.user.loaded)
        self.assertTrue(self.user.info_loaded)
        self.assertEqual(len(self.user.all_replays()), 3)
        self.assertEqual(len(self.user[:]), 3)
        self.cg.load(self.user)
        self.assertTrue(self.user.loaded)
        self.assertTrue(self.user.info_loaded)

    def test_user_slice(self):
        # sanity check (user id better be what we put in)
        self.assertEqual(self.user[0].user_id, 124493)
        # 2nd (FDFD)
        self.assertEqual(self.user[1].map_id, 129891)
        # 1st, 2nd, and 3rd (FDED, FDFD, remote control)
        self.assertListEqual([r.map_id for r in self.user[0:3]], [2249059, 129891, 774965])
        # 1st and 3rd (FDEF, remote control)
        self.assertListEqual([r.map_id for r in self.user[0:3:2]], [2249059, 774965])

    def test_no_replays_does_not_raise(self):
        u = User(12092800, "1-2", Mod.FL + Mod.EZ)
        self.cg.load_info(u)
        self.assertEqual(len(u), 0)

    def test_replay_username_loaded(self):
        # previously, usernames of `ReplayMap`s generated by `User`s would be
        # `None` due to api weirdness. Make sure this doesn't happen.
        self.assertEqual(self.tybug[0].username, "tybug")


class TestMapUser(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mu = MapUser(795627, 6304246, span="1-2")

    def test_map_user_load(self):
        self.assertEqual(len(self.mu.all_replays()), 0)
        self.assertEqual(len(self.mu[:]), 0)
        self.assertFalse(self.mu.loaded)
        self.assertFalse(self.mu.info_loaded)

        self.cg.load_info(self.mu)
        self.assertFalse(self.mu.loaded)
        self.assertTrue(self.mu.info_loaded)
        self.assertEqual(len(self.mu.all_replays()), 2)
        self.assertEqual(len(self.mu[:]), 2)

        self.cg.load(self.mu)
        self.assertTrue(self.mu.loaded)
        self.assertTrue(self.mu.info_loaded)

    def test_map_user_slice(self):
        # sanity checks (user and map id better be what we put in)
        self.assertEqual(self.mu[0].user_id, 6304246)
        self.assertEqual(self.mu[1].user_id, 6304246)
        self.assertEqual(self.mu[0].map_id, 795627)
        self.assertEqual(self.mu[1].map_id, 795627)
        # test slicing
        self.assertListEqual([r.map_id for r in self.mu[0:2]], [795627, 795627])

class TestReplayString(CGTestCase):
    def test_replay_string_load(self):
        replay_data = open(RES / "example_replay.osr", "rb").read()
        r = ReplayString(replay_data)
        self.cg.load(r)
        self.assertEqual(r.mods, Mod.HD + Mod.DT)
        self.assertEqual(r.replay_id, 2029801532)
        self.assertEqual(r.username, "MarthXT")
        self.assertEqual(r.user_id, 2909663)
        self.assertEqual(r.weight, RatelimitWeight.LIGHT)
        self.assertTrue(r.loaded)


class TestLoadableFromCG(CGTestCase):

    def test_map_from_cg(self):
        m = self.cg.Map(221777, "1-2")
        self.assertTrue(m.info_loaded)
        self.assertFalse(m.loaded)
        self.assertEqual(len(m), 2)

    def test_user_from_cg(self):
        u = self.cg.User(124493, "2, 3-4")
        self.assertTrue(u.info_loaded)
        self.assertFalse(u.loaded)
        self.assertEqual(len(u), 3)

    def test_map_user_from_cg(self):
        mu = self.cg.MapUser(795627, 6304246, "1")
        self.assertTrue(mu.info_loaded)
        self.assertFalse(mu.loaded)
        self.assertEqual(len(mu), 1)


class TestEquality(CGTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.user = User(2757689, "1-2") # toy, #1=sidetracked day, #2=View of The River Styx
        cls.user1 = User(2757689, "1-2", cache=False)
        cls.user2 = User(2757689, "1-2", mods=Mod.HT)
        cls.user3 = User(2757689, "1")

        cls.map = Map(1754777, "1-4") #sidetracked day: umbre, karthy, -duckleader-, toy
        cls.map1 = Map(1754777, "1-4", cache=False)
        cls.map2 = Map(1754777, "1-4", mods=Mod.HD)
        cls.map3 = Map(1754777, "1-2")

        cls.r = ReplayMap(1754777, 2766034) # umbre +HDHR on sidetracked day
        cls.r1 = ReplayMap(1754777, 2766034, cache=True)
        cls.r2 = ReplayMap(1754777, 2766034, mods=Mod.NF)

        cls.r3 = ReplayPath(RES / "legit" / "legit-1.osr")
        cls.r4 = ReplayPath(RES / "legit" / "legit-1.osr", cache=True)
        cls.r5 = ReplayPath(RES / "stolen_replay1.osr")

        cls.cg.load_info(cls.user)
        cls.cg.load_info(cls.user1)
        cls.cg.load_info(cls.user2)

        cls.cg.load_info(cls.map)
        cls.cg.load_info(cls.map1)
        cls.cg.load_info(cls.map2)

    def test_equality_user(self):
        self.assertEqual(self.user, self.user1)
        self.assertNotEqual(self.user, self.user2)
        self.assertNotEqual(self.user, self.user3)

    def test_equality_map(self):
        self.assertEqual(self.map, self.map1)
        self.assertNotEqual(self.map, self.map2)
        self.assertNotEqual(self.map, self.map3)

    def test_equality_replaymap(self):
        self.assertEqual(self.r, self.r1)
        self.assertNotEqual(self.r, self.r2)

    def test_equality_replaypath(self):
        self.assertEqual(self.r3, self.r4)
        self.assertNotEqual(self.r3, self.r5)
        # loaded == unloaded at the same path is true
        self.cg.load(self.r3)
        self.assertEqual(self.r3, self.r4)
        # loaded == loaded is true regardless of path
        # TODO add test where replay data in the file changes after only one
        # ReplayPath is loaded and ensure the ReplayPaths aren't equal anymore
        self.cg.load(self.r4)
        self.assertEqual(self.r3, self.r4)



import numpy as np
from slider.beatmap import Circle as SliderCircle
from slider.beatmap import Slider as SliderSlider
from slider.beatmap import Spinner as SliderSpinner
from slider.mod import circle_radius

from circleguard.mod import Mod

# We define our own hitobjects as the slider library's hitobjects have too many
# attributes and methods we don't care about, and they also lock position
# behind an extra attribute access (``hitobj.position.x``` vs ``hitobj.x``)
# which I'm not a fan of.
# Another necessary change is our hitobjects are "replay/map aware", which means
# they know how large they are (or potentially how difficult they are to hit
# as a result of OD if that is necessary in the future) and where they are (for
# HR) because we know with what mods and on what map the hitobject was played
# with.

class Hitobject:
    """
    A Hitobject in osu! gameplay, with a time and a position.
    """
    def __init__(self, time, xy):
        # TODO remove ``t`` in core 6.0.0, ``time`` should be preferred
        self.t = time
        self.time = time
        self.xy = xy
        self.x = xy[0]
        self.y = xy[1]

    @classmethod
    def from_slider_hitobj(cls, hitobj, replay, beatmap, already_converted=False):
        """
        Instantiates a circleguard hitobject from a
        :class:`slider.beatmap.HitObject`, a
        :class:`circleguard.loadables.Replay`, that the hitobject was hit on,
        and a :class:`slider.beatmap.Beatmap` that the hitobject is found in.

        The `already_converted` parameter is only to work around
        https://github.com/llllllllll/slider/issues/80 and will be removed when
        it is fixed.
        """
        easy = Mod.EZ in replay.mods
        hard_rock = Mod.HR in replay.mods
        CS = beatmap.cs(easy=easy, hard_rock=hard_rock)

        # Convert to ms.
        t = hitobj.time.total_seconds() * 1000
        # Due to floating point errors, ``t`` could actually be something
        # like ``129824.99999999999`` or ``128705.00000000001``, so round to the
        # nearest int.
        t = int(round(t))

        if hard_rock and not already_converted:
            hitobj = hitobj.hard_rock

        xy = [hitobj.position.x, hitobj.position.y]
        xy = np.array(xy)

        radius = circle_radius(CS)

        if isinstance(hitobj, SliderCircle):
            return Circle(t, xy, radius)
        if isinstance(hitobj, SliderSlider):
            return Slider(t, xy, radius)
        if isinstance(hitobj, SliderSpinner):
            return Spinner(t, xy)

    def __eq__(self, other):
        return self.time == other.time and self.xy == other.xy

    def __hash__(self):
        return hash((self.time, self.xy))


class Circle(Hitobject):
    """
    A circle in osu! gameplay, with a time, position, and radius.
    """
    def __init__(self, time, xy, radius):
        super().__init__(time, xy)
        self.radius = radius

    def __eq__(self, other):
        return (self.time == other.time and self.xy == other.xy and
            self.radius == other.radius)

    def __hash__(self):
        return hash((self.time, self.xy, self.radius))


class Slider(Hitobject):
    """
    A slider in osu! gameplay, with a time, position, and radius.
    """
    def __init__(self, time, xy, radius):
        super().__init__(time, xy)
        self.radius = radius

    def __eq__(self, other):
        return (self.time == other.time and self.xy == other.xy and
            self.radius == other.radius)

    def __hash__(self):
        return hash((self.time, self.xy, self.radius))


class Spinner(Hitobject):
    """
    A spinner in osu! gameplay, with a time and position.
    """
    def __init__(self, time, xy):
        super().__init__(time, xy)



from enum import Enum, auto

import numpy as np

from circleguard.utils import check_param
from circleguard.hitobjects import Hitobject

class JudgmentType(Enum):
    Hit300 = auto()
    Hit100 = auto()
    Hit50 = auto()
    Miss = auto()

class Judgment:
    """
    A judgment on a hitobject. A "judgment" is either a hit or a miss, with
    a hit being further classified as a 300, 100, or 50.

    Parameters
    ----------
    hitobject: :class:`slider.beatmap.HitObject`
        The hitobject being judged. This is converted to a
        :class:`circleguard.hitobjects.Hitobject`.
    replay: :class:`circleguard.loadables.Replay`
        The replay this judgment was made on.
    beatmap: :class:`slider.beatmap.Beatmap`
        The beatmap this judgment was made on.
    type: :class:`JudgmentType`
        The type of this judgment (either Hit300, Hit100, or Hit50, or Miss).
    """
    def __init__(self, hitobject, replay, beatmap, type_):
        # TODO remove `already_converted=True` when
        # https://github.com/llllllllll/slider/issues/80 is fixed
        self.hitobject = Hitobject.from_slider_hitobj(hitobject, replay,
            beatmap, True)
        self.type = type_

class Miss(Judgment):
    """
    A miss on a hitobject when a replay is played against a beatmap.
    """
    def __init__(self, hitobject, replay, beatmap):
        super().__init__(hitobject, replay, beatmap, JudgmentType.Miss)

class Hit(Judgment):
    """
    A hit on a hitobject when a replay is played against a beatmap.

    Parameters
    ----------
    hitobject: :class:`slider.beatmap.HitObject`
        The hitobject that was hit. This is converted to a
        :class:`circleguard.hitobjects.Hitobject`.
    t: float
        The time the hit occured.
    xy: list[float, float]
        The x and y position where the hit occured.
    replay: :class:`circleguard.loadables.Replay`
        The replay this hit was made on.
    beatmap: :class:`slider.beatmap.Beatmap`
        The beatmap this hit was made on.
    type: :class:`JudgmentType`
        The type of this hit (either Hit300, Hit100, or Hit50).
    """
    def __init__(self, hitobject, t, xy, replay, beatmap, type_):
        super().__init__(hitobject, replay, beatmap, type_)
        # TODO remove ``t`` in core 6.0.0, ``time`` is more intuitive. ``x`` and
        # ``y`` are fine as is though since there's no longer name for them.
        self.t = t
        self.time = t
        self.xy = xy
        self.x = xy[0]
        self.y = xy[1]
        self.type = type_

    def distance(self, *, to):
        """
        The distance from this hit to either the center or edge of its
        hitobject.

        Parameters
        ----------
        to: {"center", "edge"}
            If ``center``, the distance from this hit to the center of its
            hitobject is calculated. If ``edge``, the distance from this hit to
            the edge of its hitobject is calculated.
        Returns
        -------
        float
            The distance from this hit to either the center or edge of its
            hitobject.
        """
        check_param(to, ["center", "edge"])

        hitobj_xy = self.hitobject.xy

        if to == "edge":
            dist = np.linalg.norm(self.xy - hitobj_xy) - self.hitobject.radius
            # value is negative since we're inside the hitobject, so take abs
            return abs(dist)

        if to == "center":
            return np.linalg.norm(self.xy - hitobj_xy)

    def within(self, distance):
        """
        Whether the hit was within ``distance`` of the edge of its hitobject.

        Parameters
        ----------
        distance: float
            How close, in pixels, to the edge of the hitobject the hit has to
            be.

        Returns
        -------
        bool
            Whether the hit was within ``distance`` of the edge of its
            hitobject.

        Notes
        -----
        The lower the value, the closer to the edge the hit occurred. This value
        can never be greater than the radius of the hitobject.
        """

        return self.distance(to="edge") < distance

    def error(self):
        """
        How many milliseconds off this hit was from being a perfectly on time
        hit. If negative, this was an early hit. If positive, this was a late
        hit. If 0, this was a perfect hit.

        Returns
        -------
        float
            How many milliseconds off this hit was from being perfectly on time.
        """
        return self.time - self.hitobject.time

    def __eq__(self, other):
        return (self.hitobject == other.hitobject and self.t == other.t and
            self.xy == other.xy)

    def __hash__(self):
        return hash((self.hitobject, self.t, self.xy))

    def __repr__(self):
        return f"Hit(hitobject={self.hitobject},t={self.t},xy={self.xy}"

    def __str__(self):
        return f"({self.x}, {self.y}) at t {self.t}"
